#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Entrypoint for making requests.
"""

import copy
import urllib.parse
import pathlib
from meerschaum.utils.typing import Any, Optional, Dict, Union
from meerschaum.utils.debug import dprint
from meerschaum.utils.formatting import pprint

METHODS = {
    'GET',
    'OPTIONS',
    'HEAD',
    'POST',
    'PUT',
    'PATCH',
    'DELETE',
}

def make_request(
        self,
        method: str,
        r_url: str,
        headers: Optional[Dict[str, Any]] = None,
        use_token: bool = True,
        debug: bool = False,
        **kwargs: Any
    ) -> 'requests.Response':
    """
    Make a request to this APIConnector's endpoint using the in-memory session.

    Parameters
    ----------
    method: str
        The kind of request to make.
        Accepted values:
        - `'GET'`
        - `'OPTIONS'`
        - `'HEAD'`
        - `'POST'`
        - `'PUT'`
        - `'PATCH'`
        - `'DELETE'`

    r_url: str
        The relative URL for the endpoint (e.g. `'/pipes'`).

    headers: Optional[Dict[str, Any]], default None
        The headers to use for the request.
        If `use_token` is `True`, the authorization token will be added to a copy of these headers.

    use_token: bool, default True
        If `True`, add the authorization token to the headers.

    debug: bool, default False
        Verbosity toggle.

    kwargs: Any
        All other keyword arguments are passed to `requests.request`.

    Returns
    -------
    A `requests.Reponse` object.
    """
    if method.upper() not in METHODS:
        raise ValueError(f"Method '{method}' is not supported.")

    verify = self.__dict__.get('verify', None)
    if 'verify' not in kwargs and isinstance(verify, bool):
        kwargs['verify'] = verify

    headers = (
        copy.deepcopy(headers)
        if isinstance(headers, dict)
        else {}
    )

    if use_token:
        headers.update({'Authorization': f'Bearer {self.token}'})

    request_url = urllib.parse.urljoin(self.url, r_url)
    if debug:
        dprint(f"[{self}] Sending a '{method.upper()}' request to {request_url}")

    return self.session.request(
        method.upper(),
        request_url,
        headers = headers,
        **kwargs
    )


def get(self, r_url: str, **kwargs: Any) -> 'requests.Response':
    """
    Wrapper for `requests.get`.

    Parameters
    ----------
    r_url: str
        The relative URL for the endpoint (e.g. `'/pipes'`).

    headers: Optional[Dict[str, Any]], default None
        The headers to use for the request.
        If `use_token` is `True`, the authorization token will be added to a copy of these headers.

    use_token: bool, default True
        If `True`, add the authorization token to the headers.

    debug: bool, default False
        Verbosity toggle.

    kwargs: Any
        All other keyword arguments are passed to `requests.request`.

    Returns
    -------
    A `requests.Reponse` object.

    """
    return self.make_request('GET', r_url, **kwargs)


def post(self, r_url: str, **kwargs: Any) -> 'requests.Response':
    """
    Wrapper for `requests.post`.

    Parameters
    ----------
    r_url: str
        The relative URL for the endpoint (e.g. `'/pipes'`).

    headers: Optional[Dict[str, Any]], default None
        The headers to use for the request.
        If `use_token` is `True`, the authorization token will be added to a copy of these headers.

    use_token: bool, default True
        If `True`, add the authorization token to the headers.

    debug: bool, default False
        Verbosity toggle.

    kwargs: Any
        All other keyword arguments are passed to `requests.request`.

    Returns
    -------
    A `requests.Reponse` object.

    """
    return self.make_request('POST', r_url, **kwargs)


def patch(self, r_url: str, **kwargs: Any) -> 'requests.Response':
    """
    Wrapper for `requests.patch`.

    Parameters
    ----------
    r_url: str
        The relative URL for the endpoint (e.g. `'/pipes'`).

    headers: Optional[Dict[str, Any]], default None
        The headers to use for the request.
        If `use_token` is `True`, the authorization token will be added to a copy of these headers.

    use_token: bool, default True
        If `True`, add the authorization token to the headers.

    debug: bool, default False
        Verbosity toggle.

    kwargs: Any
        All other keyword arguments are passed to `requests.request`.

    Returns
    -------
    A `requests.Reponse` object.
    """
    return self.make_request('PATCH', r_url, **kwargs)


def put(self, r_url: str, **kwargs: Any) -> 'requests.Response':
    """
    Wrapper for `requests.put`.

    Parameters
    ----------
    r_url: str
        The relative URL for the endpoint (e.g. `'/pipes'`).

    headers: Optional[Dict[str, Any]], default None
        The headers to use for the request.
        If `use_token` is `True`, the authorization token will be added to a copy of these headers.

    use_token: bool, default True
        If `True`, add the authorization token to the headers.

    debug: bool, default False
        Verbosity toggle.

    kwargs: Any
        All other keyword arguments are passed to `requests.request`.

    Returns
    -------
    A `requests.Reponse` object.
    """
    return self.make_request('PUT', r_url, **kwargs)


def delete(self, r_url: str, **kwargs: Any) -> 'requests.Response':
    """
    Wrapper for `requests.delete`.

    Parameters
    ----------
    r_url: str
        The relative URL for the endpoint (e.g. `'/pipes'`).

    headers: Optional[Dict[str, Any]], default None
        The headers to use for the request.
        If `use_token` is `True`, the authorization token will be added to a copy of these headers.

    use_token: bool, default True
        If `True`, add the authorization token to the headers.

    debug: bool, default False
        Verbosity toggle.

    kwargs: Any
        All other keyword arguments are passed to `requests.request`.

    Returns
    -------
    A `requests.Reponse` object.
    """
    return self.make_request('DELETE', r_url, **kwargs)


def wget(
        self,
        r_url: str,
        dest: Optional[Union[str, pathlib.Path]] = None,
        headers: Optional[Dict[str, Any]] = None,
        use_token: bool = True,
        debug: bool = False,
        **kw: Any
    ) -> pathlib.Path:
    """Mimic wget with requests.
    """
    from meerschaum.utils.misc import wget
    if headers is None:
        headers = {}
    if use_token:
        headers.update({'Authorization': f'Bearer {self.token}'})
    request_url = urllib.parse.urljoin(self.url, r_url)
    if debug:
        dprint(
            f"[{self}] Downloading {request_url}"
            + (f' to {dest}' if dest is not None else '')
            + "..."
        )
    return wget(request_url, dest=dest, headers=headers, **kw)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Synchronize across config files
"""

from __future__ import annotations
from meerschaum.utils.typing import Optional, List, Tuple

def sync_yaml_configs(
        keys: List[str],
        sub_path: pathlib.Path,
        substitute: bool = True,
        permissions: Optional[int] = None,
        replace_tuples: Optional[List[Tuple[str, str]]] = None,
    ) -> None:
    """Synchronize sub-configuration with main configuration file.
    
    Parameters
    ----------
    keys: List[str]
        The config keys to read via `get_config()`.

    sub_path: pathlib.Path
        The derivative file to write.

    substitute: bool, default True
        If `True`, parse `MRSM{}` syntax and substitute values.
        See `get_config()` for more information.

    permissions: Optional[int], default None
        If not `None`, set permissions of the derivative file.

    replace_tuples: Optional[List[Tuple[str]]], default None
        If provided, iterate through a list of tuples,
        replacing the old string (index 0) with the new string (index 1).
    """
    import os, sys
    try:
        from meerschaum.utils.yaml import yaml, _yaml
    except Exception as e:
        return
    from meerschaum.config._patch import apply_patch_to_config
    import meerschaum.config
    from meerschaum.utils.packages import reload_package
    from meerschaum.config._read_config import search_and_substitute_config, get_keyfile_path

    def _read_yaml_config(path):
        """Read YAML file with header comment."""
        if not path.exists():
            return "", {}
        header_comment = ""
        with open(path, 'r') as f:
            if _yaml is not None:
                config = yaml.load(f)
            else:
                print("PyYAML not installed!")
                sys.exit(1)
            f.seek(0)
            for line in f:
                if not line.startswith('#') and not line == '\n':
                    break
                header_comment += line
        return header_comment, config

    from meerschaum.config import get_config
    config_path = get_keyfile_path(keys[0], create_new=True)
    c = get_config(*keys, substitute=substitute, sync_files=False)

    config_header, config = _read_yaml_config(config_path)
    sub_header, sub_config = _read_yaml_config(sub_path)

    if substitute:
        sub_config = search_and_substitute_config(sub_config)

    sub_config = c
    new_config_text = yaml.dump(c, sort_keys=False)
    if replace_tuples:
        for replace_tuple in replace_tuples:
            new_config_text = new_config_text.replace(replace_tuple[0], replace_tuple[1])
    new_header = sub_header
    new_path = sub_path

    ### write changes
    with open(new_path, 'w+') as f:
        f.write(new_header)
        f.write(new_config_text)
    if permissions is not None:
        os.chmod(new_path, permissions)

def sync_files(keys: Optional[List[str]] = None):
    if keys is None:
        keys = []

    def _stack():
        import os
        from meerschaum.config._paths import (
            CONFIG_DIR_PATH,
            STACK_ENV_PATH,
            STACK_COMPOSE_PATH,
            STACK_COMPOSE_FILENAME,
            STACK_ENV_FILENAME,
            GRAFANA_DATASOURCE_PATH,
            GRAFANA_DASHBOARD_PATH,
        )
        from meerschaum.config.static import STATIC_CONFIG

        sync_yaml_configs(
            ['stack', STACK_COMPOSE_FILENAME],
            STACK_COMPOSE_PATH,
            substitute = True,
            replace_tuples = [('$', '$$'), ('<DOLLAR>', '$')],
        )
        sync_yaml_configs(
            ['stack', 'grafana', 'datasource'],
            GRAFANA_DATASOURCE_PATH,
            substitute = True,
        )
        sync_yaml_configs(
            ['stack', 'grafana', 'dashboard'],
            GRAFANA_DASHBOARD_PATH,
            substitute = True,
        )


    key_functions = {
        'stack': _stack,
    }
    if keys is None:
        keys = list(key_functions.keys())
    for k in keys:
        if k in key_functions:
            key_functions[k]()




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Install plugins and pip packages.
"""

from __future__ import annotations
from meerschaum.utils.typing import List, Any, SuccessTuple, Optional, Union

def install(
        action: Optional[List[str]] = None,
        **kw: Any
    ) -> SuccessTuple:
    """
    Install Meerschaum plugins or Python packages.
    """
    from meerschaum.actions import choose_subaction
    options = {
        'plugins'  : _install_plugins,
        'packages' : _install_packages,
        'required' : _install_required,
    }
    return choose_subaction(action, options, **kw)


def _complete_install(
        action: Optional[List[str]] = None,
        **kw: Any
    ) -> List[str]:
    """
    Override the default Meerschaum `complete_` function.
    """
    if action is None:
        action = []
    options = {
        'plugin' : _complete_install_plugins,
        'plugins' : _complete_install_plugins,
        'package': _complete_install_packages,
        'packages': _complete_install_packages,
        'required': _complete_install_required,
    }

    if len(action) > 0 and action[0] in options:
        sub = action[0]
        del action[0]
        return options[sub](action=action, **kw)

    from meerschaum._internal.shell import default_action_completer
    return default_action_completer(action=(['install'] + action), **kw)


def _install_plugins(
        action: Optional[List[str]] = None,
        repository: Optional[str] = None,
        skip_deps: bool = False,
        force: bool = False,
        debug: bool = False,
        **kw: Any
    ) -> SuccessTuple:
    """
    Install a plugin.
    
    By default, install from the main Meerschaum repository (mrsm.io).
    Use a private repository by specifying the API label after the plugin.
    NOTE: the --instance flag is ignored!
    
    Usage:
        install plugins [plugin]
    
    Examples:
        - install plugins noaa
        - install plugins noaa --repo mrsm  (mrsm is the default instance)
        - install plugins noaa --repo mycustominstance
    """
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.warnings import info
    from meerschaum.plugins import reload_plugins
    from meerschaum.connectors.parse import parse_repo_keys
    import meerschaum.actions
    from meerschaum.utils.formatting import print_tuple
    from meerschaum.core import Plugin
    from meerschaum.connectors.api import APIConnector

    if not action:
        return False, "No plugins to install."

    repo_connector = parse_repo_keys(repository)

    for name in action:
        info(f"Installing plugin '{name}' from Meerschaum repository '{repo_connector}'...")
        success, msg = repo_connector.install_plugin(
            name,
            force = force,
            skip_deps = skip_deps,
            debug = debug,
        )
        print_tuple((success, msg))

    reload_plugins(debug=debug)
    return True, "Success"


def _complete_install_plugins(
        action: Optional[List[str]] = None,
        repository: Optional[str] = None,
        **kw: Any
    ) -> List[str]:
    """
    Search for plugins to autocomplete command line text.
    NOTE: Disabled for the time being so we don't interrupt the user typing.
    """
    return []
    from meerschaum.actions import get_shell
    if action is None:
        action = []
    if len(action) == 0:
        search_term = None
    else:
        search_term = action[-1]

    ### Don't start searching unless a key has been pressed.
    if search_term is None or len(search_term) == 0:
        return []

    ### In case we're using the Shell (which we have to in order for _complete to work),
    ### get the current repository.
    if repository is None:
        repository = get_shell().repo_keys

    from meerschaum.connectors.parse import parse_repo_keys
    try:
        repo_connector = parse_repo_keys(repository)
    except Exception as e:
        return []
    results = repo_connector.get_plugins(search_term=search_term)
    if len(results) == 1 and results[0] == search_term:
        return []
    return sorted(results)

class NoVenv:
    pass

def _install_packages(
        action: Optional[List[str]] = None,
        sub_args: Optional[List[str]] = None,
        venv: Union[str, NoVenv] = NoVenv,
        debug: bool = False,
        **kw: Any
    ) -> SuccessTuple:
    """
    Install PyPI packages into the Meerschaum virtual environment.
    
    Example:
        `install packages pandas numpy`
    """
    if not action:
        return False, f"No packages to install."
    from meerschaum.utils.warnings import info
    from meerschaum.utils.packages import pip_install
    from meerschaum.utils.misc import items_str
    if venv is NoVenv:
        venv = 'mrsm'

    if pip_install(
        *action,
        args = ['--upgrade'] + sub_args,
        venv = venv,
        debug = debug,
    ):
        return True, (
            "Successfully installed package" + ("s" if len(action) != 1 else '')
            + f" {items_str(action)}"
            + f" into the virtual environment '{venv}'."
        )
    return False, (
        f"Failed to install package" + ("s" if len(action) != 1 else '') + f" {items_str(action)}."
    )


def _complete_install_packages(
        **kw : Any
    ) -> List[str]:
    return []


def _install_required(
        action: Optional[List[str]] = None,
        repository: Optional[str] = None,
        force: bool = False,
        debug: bool = False,
        **kw
    ) -> SuccessTuple:
    """
    Install the required packages for Meerschaum plugins.
    Each plugin's packages will be installed into its virtual environment.

    Example:
        `install required noaa covid`
    """
    from meerschaum.core import Plugin
    from meerschaum.utils.warnings import warn, info
    from meerschaum.connectors.parse import parse_repo_keys
    from meerschaum.utils.formatting import print_tuple
    from meerschaum.plugins import get_plugins_names
    repo_connector = parse_repo_keys(repository)

    plugins_names = action or get_plugins_names()

    success_count = 0
    fail_count = 0

    for plugin_name in plugins_names:
        plugin = Plugin(plugin_name, repo_connector=repo_connector)
        if not plugin.is_installed():
            warn(f"Plugin '{plugin}' is not installed. Skipping...", stack=False)
            continue
        info(f"Installing required packages for plugin '{plugin}'...")
        success = plugin.install_dependencies(force=force, debug=debug)
        if not success:
            warn(f"Failed to install required packages for plugin '{plugin}'.", stack=False)
            fail_count += 1
        else:
            success_count += 1

    success = fail_count == 0
    msg = (
        f"Installed packages for {success_count + fail_count} plugins\n    "
        + f"({success_count} succeeded, {fail_count} failed)."
    )
    return success, msg


def _complete_install_required(*args, **kw) -> List[str]:
    from meerschaum.actions.uninstall import _complete_uninstall_plugins
    return _complete_uninstall_plugins(*args, **kw)


### NOTE: This must be the final statement of the module.
###       Any subactions added below these lines will not
###       be added to the `help` docstring.
from meerschaum.actions import choices_docstring as _choices_docstring
install.__doc__ += _choices_docstring('install')



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

import pytest
from datetime import datetime, timedelta
from decimal import Decimal
from tests import debug
from tests.pipes import all_pipes, stress_pipes, remote_pipes
from tests.connectors import conns, get_flavors
from tests.test_users import test_register_user
import meerschaum as mrsm
from meerschaum import Pipe
from meerschaum.actions import actions

@pytest.fixture(autouse=True)
def run_before_and_after(flavor: str):
    test_register_user(flavor)
    yield


@pytest.mark.parametrize("flavor", get_flavors())
def test_register_and_delete(flavor: str):
    pipe = all_pipes[flavor][0]
    params = pipe.parameters.copy()
    assert params is not None
    output = pipe.delete()
    pipe.parameters = params
    assert pipe.parameters is not None
    success, msg = pipe.register(debug=debug)
    assert success, msg
    assert pipe.get_id(debug=debug) is not None
    success, msg = pipe.delete(debug=debug)
    assert success, msg
    pipe.parameters = params
    success, msg = pipe.register(debug=debug)
    assert success, msg
    assert pipe.parameters is not None

@pytest.mark.parametrize("flavor", get_flavors())
def test_drop_and_sync(flavor: str):
    pipe = all_pipes[flavor][0]
    pipe.drop()
    assert pipe.exists(debug=debug) is False
    assert pipe.columns is not None
    now1 = datetime(2021, 1, 1, 12, 0)
    data = {'datetime' : [now1], 'id' : [1], 'val': [1]}
    success, msg = pipe.sync(data, debug=debug)
    assert success, msg
    assert pipe.exists(debug=debug)
    now2 = datetime(2021, 1, 1, 12, 1)
    data = {'datetime' : [now2], 'id' : [1], 'val': [1]}
    success, msg = pipe.sync(data, debug=debug)
    assert success, msg
    assert pipe.exists(debug=debug)
    data = pipe.get_data(debug=debug)
    assert data is not None
    assert len(data) == 2

@pytest.mark.parametrize("flavor", get_flavors())
def test_drop_and_sync_duplicate(flavor: str):
    """
    Verify dropping a table and syncing duplicate rows are filtered out.
    """
    pipe = all_pipes[flavor][0]
    pipe.drop(debug=debug)
    assert not pipe.exists(debug=debug)

    now1 = datetime(2021, 1, 1, 12, 0)
    data = {'datetime': [now1], 'id': [1], 'val': [1]}
    success, msg = pipe.sync(data, debug=debug)
    assert success, msg
    data = pipe.get_data(debug=debug)
    assert len(data) == 1

    now1 = datetime(2021, 1, 1, 12, 0)
    data = {'datetime': [now1], 'id': [1], 'val': [1]}
    success, msg = pipe.sync(data, debug=debug)
    assert success, msg
    data = pipe.get_data(debug=debug)
    assert len(data) == 1

@pytest.mark.parametrize("flavor", get_flavors())
def test_drop_and_sync_stress(flavor: str):
    pipe = stress_pipes[flavor]
    pipe.drop(debug=debug)
    success, msg = pipe.sync(debug=debug)
    assert success, msg

@pytest.mark.parametrize("flavor", get_flavors())
def test_drop_and_sync_remote(flavor: str):
    pipe = None
    for p in remote_pipes[flavor]:
        if str(p.connector) == str(p.instance_connector):
            pipe = p
            break
    assert pipe is not None
    pipe.delete(debug=debug)
    parent_pipe = Pipe('plugin:stress', 'test', instance=pipe.connector)
    parent_pipe.delete(debug=debug)
    begin, end = datetime(2020, 1, 1), datetime(2020, 1, 2)
    success, msg = parent_pipe.sync(begin=begin, end=end, debug=debug)
    parent_len = parent_pipe.get_rowcount(debug=debug)
    assert success, msg

    success, msg = pipe.sync(debug=debug)
    assert success, msg
    child_len = pipe.get_rowcount(debug=debug)
    assert parent_len == child_len

    success, msg = parent_pipe.sync(
        [{'datetime': '2020-01-03', 'id': -1, 'foo': 'a'}],
        debug = debug,
    )
    assert success, msg
    parent_len2 = parent_pipe.get_rowcount(debug=debug)
    assert parent_len2 == (parent_len + 1)
    success, msg = pipe.sync(debug=debug)
    assert len(pipe.get_columns_types(debug=debug)) == 4
    child_len2 = pipe.get_rowcount(debug=debug)
    assert parent_len2 == child_len2
    success, msg = parent_pipe.sync(
        [{'datetime': '2020-01-03', 'id': -1, 'foo': 'b'}],
        debug = debug,
    )
    assert success, msg
    parent_len3 = parent_pipe.get_rowcount(debug=debug)
    assert parent_len2 == parent_len3
    success, msg = pipe.sync(debug=debug, begin='2020-01-01')
    assert success, msg
    child_len3 = pipe.get_rowcount(debug=debug)
    assert child_len3 == parent_len3
    df = pipe.get_data(params={'id': -1}, debug=debug)
    assert len(df) == 1
    assert df.to_dict(orient='records')[0]['foo'] == 'b'


@pytest.mark.parametrize("flavor", get_flavors())
def test_sync_engine(flavor: str):
    ### Weird concurrency issues with our tests.
    if flavor == 'duckdb':
        return
    pipe = stress_pipes[flavor]
    mrsm_instance = str(pipe.instance_connector)
    success, msg = actions['drop'](
        ['pipes'],
        connector_keys = [pipe.connector_keys],
        metric_keys = [pipe.metric_key],
        location_keys = [pipe.location_key],
        mrsm_instance = mrsm_instance,
        yes = True,
        #  debug = True,
    )
    assert success, msg

    success, msg = actions['sync'](
        ['pipes'],
        connector_keys = [pipe.connector_keys],
        metric_keys = [pipe.metric_key],
        location_keys = [pipe.location_key],
        mrsm_instance = mrsm_instance,
        #  debug = True,
    )
    assert success, msg


@pytest.mark.parametrize("flavor", get_flavors())
def test_target_mutable(flavor: str):
    conn = conns[flavor]
    if conn.type != 'sql':
        return
    target = 'FooBar!'
    pipe = Pipe('foo', 'bar', target=target, instance=conn, columns={'datetime': 'dt', 'id': 'id'})
    pipe.drop(debug=debug)
    assert not pipe.exists(debug=debug)
    success, msg = pipe.sync(
        {'dt': [datetime(2022, 6, 8)], 'id': [1], 'vl': [10]},
        debug = debug
    )
    df = conn.read(target)
    assert len(df) == 1
    success, msg = pipe.sync(
        {'dt': [datetime(2022, 6, 8)], 'id': [1], 'vl': [10]},
        debug = debug
    )
    df = conn.read(target)
    assert len(df) == 1
    success, msg = pipe.sync(
        {'dt': [datetime(2022, 6, 8)], 'id': [1], 'vl': [100]},
        debug = debug
    )
    df = conn.read(target)
    assert len(df) == 1
    pipe.drop()
    result = conn.read(target, silent=True)
    assert result is None


@pytest.mark.parametrize("flavor", get_flavors())
def test_sync_new_columns(flavor: str):
    """
    Test that new columns are added.
    """
    conn = conns[flavor]
    pipe = Pipe('foo', 'bar', columns={'datetime': 'dt', 'id': 'id'}, instance=conn)
    pipe.drop(debug=debug)
    docs = [
        {'dt': '2022-01-01', 'id': 1, 'a': 10},
    ]
    pipe.sync(docs, debug=debug)
    assert len(pipe.get_data().columns) == 3

    docs = [
        {'dt': '2022-01-01', 'id': 1, 'b': 20},
    ]
    pipe.sync(docs, debug=debug)
    df = pipe.get_data()
    assert len(df.columns) == 4
    assert len(df) == 1


@pytest.mark.parametrize("flavor", get_flavors())
def test_temporary_pipes(flavor: str):
    """
    Verify that `temporary=True` will not create instance tables.
    """
    from meerschaum.utils.misc import generate_password
    from meerschaum.utils.sql import table_exists
    if flavor != 'sqlite':
        return
    session_id = generate_password(6)
    db_path = '/tmp/' + session_id + '.db'
    conn = mrsm.get_connector('sql', session_id, flavor='sqlite', database=db_path) 
    pipe = Pipe('foo', 'bar', instance=conn, temporary=True, columns={'id': 'id'})
    _ = pipe.parameters
    _ = pipe.id
    _ = pipe.get_rowcount(debug=debug)
    success, msg = pipe.sync([{'id': 1, 'a': 2}], debug=debug)
    assert success, msg
    success, msg = pipe.sync([{'id': 2, 'b': 3}], debug=debug)
    assert success, msg
    success, msg = pipe.sync(
        [
            {'id': 1, 'b': 4},
            {'id': 2, 'a': 5},
        ],
        debug = debug,
    )
    success, msg = pipe.delete(debug=debug)
    assert (not success), msg
    assert pipe.get_rowcount(debug=debug) == 2
    assert not table_exists('pipes', conn, debug=debug)
    assert not table_exists('users', conn, debug=debug)
    assert not table_exists('plugins', conn, debug=debug)


@pytest.mark.parametrize("flavor", get_flavors())
def test_id_index_col(flavor: str):
    """
    Verify that the ID column is able to be synced.
    """
    conn = conns[flavor]
    pipe = Pipe(
        'test_id', 'index_col', 'table',
        instance = conn,
        dtypes = {'id': 'Int64'},
        columns = {'datetime': 'id'},
    )
    pipe.delete()
    docs = [{'id': i, 'a': i*2, 'b': {'c': i/2}} for i in range(100)]
    success, msg = pipe.sync(docs, debug=debug)
    assert success, msg
    df = pipe.get_data(debug=debug)
    synced_docs = df.to_dict(orient='records')
    assert len(df) == len(docs)
    assert synced_docs == docs

    ### Sync the same docs again and verify nothing has changed.
    success, msg = pipe.sync(docs, debug=debug)
    assert success, msg
    df = pipe.get_data(debug=debug)
    synced_docs = df.to_dict(orient='records')
    assert synced_docs == docs

    ### Update the first 10 docs.
    new_docs = [{'id': i, 'a': i*3, 'b': {'c': round(i/3, 2)}} for i in range(10)]
    success, msg = pipe.sync(new_docs, debug=debug)
    small_df = pipe.get_data(end=10, debug=debug)
    print(small_df)
    assert len(small_df) == len(new_docs)
    small_synced_docs = small_df.to_dict(orient='records')
    assert small_synced_docs == new_docs


@pytest.mark.parametrize("flavor", get_flavors())
def test_utc_offset_datetimes(flavor: str):
    """
    Verify that we are able to sync rows with UTC offset datetimes.
    """
    conn = conns[flavor]
    pipe = Pipe(
        'test_utc_offset', 'datetimes',
        instance = conn,
        columns = {'datetime': 'dt'},
    )
    pipe.delete()

    docs = [
        {'dt': '2023-01-01 00:00:00+00:00'},
        {'dt': '2023-01-02 00:00:00+01:00'},
    ]

    expected_docs = [
        {'dt': datetime(2023, 1, 1)},
        {'dt': datetime(2023, 1, 1, 23, 0, 0)}
    ]

    success, msg = pipe.sync(docs, debug=debug)
    assert success, msg
    df = pipe.get_data(debug=debug)
    synced_docs = df.to_dict(orient='records')
    assert synced_docs == expected_docs


@pytest.mark.parametrize("flavor", get_flavors())
def test_ignore_datetime_conversion(flavor: str):
    """
    If the user specifies, skip columns from being detected as datetimes.
    """
    conn = conns[flavor]
    pipe = Pipe(
        'test_utc_offset', 'datetimes', 'ignore',
        instance = conn,
        dtypes = {
            'dt': 'str',
        },
    )
    pipe.delete()

    docs = [
        {'dt': '2023-01-01 00:00:00+00:00'},
        {'dt': '2023-01-02 00:00:00+01:00'},
    ]

    expected_docs = [
        {'dt': '2023-01-01 00:00:00+00:00'},
        {'dt': '2023-01-02 00:00:00+01:00'},
    ]

    success, msg = pipe.sync(docs, debug=debug)
    assert success, msg
    df = pipe.get_data(debug=debug)
    synced_docs = df.to_dict(orient='records')
    assert synced_docs == expected_docs


@pytest.mark.parametrize("flavor", get_flavors())
def test_no_indices_inferred_datetime_to_text(flavor: str):
    """
    Verify that changing dtypes are handled.
    """
    conn = conns[flavor]
    pipe = Pipe(
        'test_no_indices', 'datetimes', 'text',
        instance = conn,
    )
    pipe.delete()
    docs = [
        {'fake-dt': '2023-01-01', 'a': 1},
    ]
    success, msg = pipe.sync(docs, debug=debug)
    assert success, msg

    docs = [
        {'fake-dt': '2023-01-01', 'a': 1},
        {'fake-dt': '2023-01-02', 'a': 2},
    ]
    success, msg = pipe.sync(docs, debug=debug)
    assert success, msg
    df = pipe.get_data()
    assert len(df) == len(docs)

    docs = [
        {'fake-dt': '2023-01-01', 'a': 1},
        {'fake-dt': '2023-01-02', 'a': 2},
        {'fake-dt': 'foo', 'a': 3},
    ]
    success, msg = pipe.sync(docs, debug=debug)
    assert success, msg
    df = pipe.get_data()
    assert len(df) == len(docs)


@pytest.mark.parametrize("flavor", get_flavors())
def test_sync_generators(flavor: str):
    """
    Verify that we are able to sync generators of chunks.
    """
    conn = conns[flavor]
    pipe = Pipe(
        'test_generators', 'foo',
        instance = conn,
        columns = {'datetime': 'dt'},
    )
    pipe.delete()
    start_time = datetime(2023, 1, 1)
    num_docs = 3
    generator = ([{'dt': start_time + timedelta(days=i)}] for i in range(num_docs))
    success, msg = pipe.sync(generator, debug=debug)
    assert success, msg
    rowcount = pipe.get_rowcount()
    assert rowcount == num_docs


@pytest.mark.parametrize("flavor", get_flavors())
def test_add_new_columns(flavor: str):
    """
    Verify that we are able to add new columns dynamically.
    """
    conn = conns[flavor]
    pipe = Pipe('test_add', 'columns', instance=conn)
    pipe.delete()
    pipe = Pipe(
        'test_add', 'columns',
        instance = conn,
        columns = {'datetime': 'dt'},
    )

    docs = [{'dt': '2023-01-01', 'a': 1}]
    success, msg = pipe.sync(docs, debug=debug)
    assert success, msg
    docs = [{'dt': '2023-01-01', 'b': 'foo', 'c': 12.3, 'd': {'e': 5}}]
    success, msg = pipe.sync(docs, debug=debug)
    assert success, msg
    assert len(pipe.dtypes) == 5
    df = pipe.get_data()
    assert len(df) == 1
    assert 'dt' in df.columns
    assert 'a' in df.columns
    assert 'b' in df.columns
    assert 'c' in df.columns
    assert 'd' in df.columns
    assert 'e' in df['d'][0]


@pytest.mark.parametrize("flavor", get_flavors())
def test_get_data_iterator(flavor: str):
    """
    Test the new `as_iterator` flag in `Pipe.get_data()`.
    """
    from meerschaum.utils.packages import import_pandas
    pd = import_pandas()
    conn = conns[flavor]
    pipe = mrsm.Pipe('test', 'get_data_iterator', instance=conn)
    pipe.delete()
    pipe = mrsm.Pipe(
        'test', 'get_data_iterator', 'foo', 
        instance = conn,
        columns = {'datetime': 'id'},
        dtypes = {'id': 'Int64'},
    )
    docs = [{'id': i, 'color': ('a' if i % 2 == 0 else 'b')} for i in range(7)]
    success, message = pipe.sync(docs, debug=debug)
    assert success, message
    gen = pipe.get_data(
        begin = 1,
        end = 6,
        as_iterator = True,
        chunk_interval = 2,
        debug = debug,
    )
    chunks = [chunk for chunk in gen]
    assert len(chunks) == 3
    assert len(chunks[2]) == 1

    gen = pipe.get_data(
        params = {'color': 'a'},
        as_iterator = True,
        chunk_interval = 3,
        debug = debug,
    )
    chunks = [chunk for chunk in gen]
    print(chunks)
    for c in chunks:
        print(c)
    df = pd.concat(chunks)
    assert list(df['id']) == [0, 2, 4, 6]


@pytest.mark.parametrize("flavor", get_flavors())
def test_sync_inplace(flavor: str):
    """
    Verify that in-place syncing works as expected.
    """
    from meerschaum.utils.sql import sql_item_name
    if flavor == 'api':
        return
    conn = conns[flavor]
    source_pipe = Pipe('test', 'inplace', 'src', instance=conn)
    source_pipe.delete()
    source_pipe = Pipe(
        'test', 'inplace', 'src',
        instance = conn,
        columns = {'datetime': 'dt'}
    )
    dest_pipe = Pipe(str(conn), 'inplace', 'dest', instance=conn)
    dest_pipe.delete()
    query = f"SELECT * FROM {sql_item_name(source_pipe.target, flavor)}"
    dest_pipe = Pipe(
        str(conn), 'inplace', 'dest',
        instance = conn,
        columns = {'datetime': 'dt'},
        parameters = {
            "fetch": {
                "definition": query,
                "backtrack_minutes": 1440,
            }
        },
    )

    docs = [
        {'dt': '2023-01-01 00:00:00'},
        {'dt': '2023-01-01 00:01:00'},
        {'dt': '2023-01-01 00:02:00'},
        {'dt': '2023-01-01 00:03:00'},
        {'dt': '2023-01-01 00:04:00'},
    ]
    success, msg = source_pipe.sync(docs)
    assert success, msg

    success, msg = dest_pipe.sync(debug=debug)
    assert success, msg
    assert dest_pipe.get_rowcount(debug=debug) == len(docs)

    success, msg = dest_pipe.sync(debug=debug)
    assert success, msg
    assert dest_pipe.get_rowcount(debug=debug) == len(docs)

    new_docs = [
        {'dt': '2023-01-02 00:00:00'},
        {'dt': '2023-01-02 00:01:00'},
        {'dt': '2023-01-02 00:02:00'},
        {'dt': '2023-01-02 00:03:00'},
        {'dt': '2023-01-02 00:04:00'},
    ]
    success, msg = source_pipe.sync(new_docs)
    assert success, msg

    success, msg = dest_pipe.sync(debug=debug)
    assert success, msg
    assert dest_pipe.get_rowcount(debug=debug) == len(docs) + len(new_docs)


@pytest.mark.parametrize("flavor", get_flavors())
def test_nested_chunks(flavor: str):
    """
    Sync nested chunk generators.
    """
    conn = conns[flavor]
    pipe = mrsm.Pipe('test', 'chunks', 'nested', instance=conn)
    pipe.delete()
    pipe = mrsm.Pipe('test', 'chunks', 'nested', instance=conn)

    num_docs = 3
    docs = [{'a': i} for i in range(num_docs)]
    chunks = ([{'chunk_ix': i, **doc}] for i, doc in enumerate(docs))
    chunks_of_chunks = (
        chunk
        for chunk in chunks
    )
    chunks_of_chunks_of_chunks = (
        g for g in chunks_of_chunks
    )
    success, msg = pipe.sync(chunks_of_chunks_of_chunks, debug=debug)
    from meerschaum.utils.formatting import print_tuple
    print_tuple((success, msg))
    assert success, msg
    df = pipe.get_data()
    assert len(df) == num_docs


@pytest.mark.parametrize("flavor", get_flavors())
def test_sync_dask_dataframe(flavor: str):
    """
    Verify that we are able to sync Dask DataFrames.
    """
    conn = conns[flavor]
    pipe = mrsm.Pipe(
        'dask', 'demo',
        columns = {'datetime': 'dt'},
        instance = conn,
    )
    pipe.drop()
    pipe.sync([
        {'dt': '2023-01-01', 'id': 1},
        {'dt': '2023-01-02', 'id': 2},
        {'dt': '2023-01-03', 'id': 3},
    ])
    ddf = pipe.get_data(as_dask=True, debug=debug)

    pipe2 = mrsm.Pipe(
        'dask', 'insert',
        columns = pipe.columns,
        instance = conn,
    )
    pipe2.drop()
    pipe2.sync(ddf, debug=debug)
    assert pipe.get_data().to_dict() == pipe2.get_data().to_dict()


@pytest.mark.parametrize("flavor", get_flavors())
def test_sync_null_indices(flavor: str):
    """
    Test that null indices are accounted for.
    """
    conn = conns[flavor]
    pipe = mrsm.Pipe('sync', 'null', 'indices', instance=conn)
    pipe.delete()
    pipe = mrsm.Pipe(
        'sync', 'null', 'indices',
        instance = conn,
        columns = ['a', 'b'],
    )
    docs = [{'a': 1, 'b': 1, 'c': 1}]
    success, msg = pipe.sync(docs, debug=debug)
    assert success, msg
    docs = [{'a': None, 'b': 1, 'c': 1}]
    success, msg = pipe.sync(docs, debug=debug)
    assert success, msg
    df = pipe.get_data(debug=debug)
    assert len(df) == 2

    docs = [{'a': 1, 'b': None, 'c': 1}]
    success, msg = pipe.sync(docs, debug=debug)
    assert success, msg
    df = pipe.get_data(debug=debug)
    assert len(df) == 3

    docs = [{'a': 1, 'b': None, 'c': 1}]
    success, msg = pipe.sync(docs, debug=debug)
    assert success, msg
    df = pipe.get_data(debug=debug)
    assert len(df) == 3

    docs = [{'a': None, 'b': None, 'c': 1}]
    success, msg = pipe.sync(docs, debug=debug)
    assert success, msg
    df = pipe.get_data(debug=debug)
    assert len(df) == 4

    docs = [{'c': Decimal('2.2')}]
    success, msg = pipe.sync(docs, debug=debug)
    assert success, msg
    df = pipe.get_data(debug=debug)
    assert len(df) == 4
    df = pipe.get_data(['c'], params={'a': None, 'b': [None]}, debug=debug)
    assert len(df) == 1
    assert df['c'][0] == Decimal('2.2')


@pytest.mark.parametrize("flavor", get_flavors())
def test_upsert_sync(flavor: str):
    """
    Test that setting `upsert` to `True` is able to sync successfully.
    """
    conn = conns[flavor]
    pipe = Pipe('test', 'upsert', instance=conn)
    pipe.delete()
    pipe = Pipe(
        'test', 'upsert', instance=conn,
        columns = {'datetime': 'dt', 'id': 'id'},
        dtypes = {'num': 'numeric'},
        parameters = {'upsert': True},
    )

    docs = [
        {'dt': '2023-01-01', 'id': 1, 'yes': True, 'num': '1.1'},
        {'dt': '2023-01-02', 'id': 2, 'yes': False, 'num': '2'},
    ]
    success, msg = pipe.sync(docs, debug=debug)
    assert success, msg

    docs = [
        {'dt': '2023-01-02', 'id': 2, 'yes': None, 'num': '2'},
    ]
    success, msg = pipe.sync(docs, debug=debug)
    assert success, msg
    df = pipe.get_data(params={'id': 2})
    assert 'na' in str(df['yes'][0]).lower()
    assert isinstance(df['num'][0], Decimal)
    assert df['num'][0] == Decimal('2')


@pytest.mark.parametrize("flavor", get_flavors())
def test_upsert_no_value_cols(flavor: str):
    """
    Test that setting `upsert` to `True` is able to sync without value columns.
    """
    conn = conns[flavor]
    pipe = Pipe('test', 'upsert', 'no_vals', instance=conn)
    pipe.delete()
    pipe = Pipe(
        'test', 'upsert', 'no_vals', instance=conn,
        columns = {'datetime': 'dt', 'id': 'id'},
        parameters = {'upsert': True},
    )

    docs = [
        {'dt': '2023-01-01', 'id': 1},
        {'dt': '2023-01-02', 'id': 2},
        {'dt': '2023-01-03', 'id': 3},
    ]
    success, msg = pipe.sync(docs, debug=debug)
    assert success, msg

    docs = [
        {'dt': '2023-01-03', 'id': 3},
    ]
    success, msg = pipe.sync(docs, debug=debug)
    assert success, msg
    return pipe
    assert pipe.get_rowcount() == 3



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Create a file-like object that manages sub-files under the hood.
"""

import os
import io
import re
import pathlib
import traceback
import sys
import atexit
from typing import List, Union, Optional, Tuple
from meerschaum.config import get_config
from meerschaum.utils.warnings import warn
import meerschaum as mrsm
daemon = mrsm.attempt_import('daemon')

class RotatingFile(io.IOBase):
    """
    A `RotatingFile` may be treated like a normal file-like object.
    Under the hood, however, it will create new sub-files and delete old ones.
    """

    SEEK_BACK_ATTEMPTS: int = 5

    def __init__(
            self,
            file_path: pathlib.Path,
            num_files_to_keep: Optional[int] = None,
            max_file_size: Optional[int] = None,
            redirect_streams: bool = False,
        ):
        """
        Create a file-like object which manages other files.

        Parameters
        ----------
        num_files_to_keep: int, default None
            How many sub-files to keep at any given time.
            Defaults to the configured value (5).

        max_file_size: int, default None
            How large in bytes each sub-file can grow before another file is created.
            Note that this is not a hard limit but rather a threshold
            which may be slightly exceeded.
            Defaults to the configured value (100_000).

        redirect_streams: bool, default False
            If `True`, redirect previous file streams when opening a new file descriptor.
            
            NOTE: Only set this to `True` if you are entering into a daemon context.
            Doing so will redirect `sys.stdout` and `sys.stderr` into the log files.
        """
        self.file_path = pathlib.Path(file_path)
        if num_files_to_keep is None:
            num_files_to_keep = get_config('jobs', 'logs', 'num_files_to_keep')
        if max_file_size is None:
            max_file_size = get_config('jobs', 'logs', 'max_file_size')
        if num_files_to_keep < 2:
            raise ValueError("At least 2 files must be kept.")
        if max_file_size < 1:
            raise ValueError("Subfiles must contain at least one byte.")

        self.num_files_to_keep = num_files_to_keep
        self.max_file_size = max_file_size
        self.redirect_streams = redirect_streams
        self.subfile_regex_pattern = re.compile(
            r'^'
            + self.file_path.name
            + r'(?:\.\d+)?$'
        )

        ### When subfiles are opened, map from their index to the file objects.
        self.subfile_objects = {}
        self._redirected_subfile_objects = {}
        self._current_file_obj = None
        self._previous_file_obj = None

        ### When reading, keep track of the file index and position.
        self._cursor: Tuple[int, int] = (0, 0)

        ### Don't forget to close any stray files.
        atexit.register(self.close)


    def fileno(self):
        """
        Return the file descriptor for the latest subfile.
        """
        self.refresh_files()
        return self._current_file_obj.fileno()


    def get_latest_subfile_path(self) -> pathlib.Path:
        """
        Return the path for the latest subfile to which to write into.
        """
        return self.get_subfile_path_from_index(
            self.get_latest_subfile_index()
        )


    def get_remaining_subfile_size(self, subfile_index: int) -> int:
        """
        Return the remaining buffer size for a subfile.

        Parameters
        ---------
        subfile_index: int
            The index of the subfile to be checked.

        Returns
        -------
        The remaining size in bytes.
        """
        subfile_path = self.get_subfile_path_from_index(subfile_index)
        if not subfile_path.exists():
            return self.max_file_size

        return self.max_file_size - os.path.getsize(subfile_path)


    def is_subfile_too_large(self, subfile_index: int, potential_new_len: int = 0) -> bool:
        """
        Return whether a given subfile is too large.

        Parameters
        ----------
        subfile_index: int
            The index of the subfile to be checked.

        potential_new_len: int, default 0
            The length of a potential write of new data.

        Returns
        -------
        A bool indicating the subfile is or will be too large.
        """
        subfile_path = self.get_subfile_path_from_index(subfile_index)
        if not subfile_path.exists():
            return False

        self.flush()

        return (
            (os.path.getsize(subfile_path) + potential_new_len)
            >=
            self.max_file_size
        )


    def get_latest_subfile_index(self) -> int:
        """
        Return the latest existing subfile index.
        If no index may be found, return -1.
        """
        existing_subfile_paths = self.get_existing_subfile_paths()
        latest_index = (
            self.get_index_from_subfile_name(existing_subfile_paths[-1].name)
            if existing_subfile_paths
            else -1
        )
        return latest_index


    def get_index_from_subfile_name(self, subfile_name: str) -> int:
        """
        Return the index from a given subfile name.
        If the file name cannot be parsed, return -1.
        """
        try:
            return int(subfile_name.replace(self.file_path.name + '.', ''))
        except Exception as e:
            return -1


    def get_subfile_name_from_index(self, subfile_index: int) -> str:
        """
        Return the subfile name from the given index.
        """
        return f'{self.file_path.name}.{subfile_index}'


    def get_subfile_path_from_index(self, subfile_index: int) -> pathlib.Path:
        """
        Return the subfile's path from its index.
        """
        return self.file_path.parent / self.get_subfile_name_from_index(subfile_index)


    def get_existing_subfile_indices(self) -> List[int]:
        """
        Return of list of subfile indices which exist on disk.
        """
        existing_subfile_paths = self.get_existing_subfile_paths()
        return [self.get_index_from_subfile_name(path.name) for path in existing_subfile_paths]


    def get_existing_subfile_paths(self) -> List[pathlib.Path]:
        """
        Return a list of file paths that match the input filename pattern.
        """
        if not self.file_path.parent.exists():
            return []

        subfile_names_indices = sorted(
            [
                (file_name, self.get_index_from_subfile_name(file_name))
                for file_name in os.listdir(self.file_path.parent)
                if re.match(self.subfile_regex_pattern, file_name)
            ],
            key = lambda x: x[1],
        )
        return [
            (self.file_path.parent / file_name)
            for file_name, _ in subfile_names_indices
        ]


    def refresh_files(self, potential_new_len: int = 0) -> '_io.TextUIWrapper':
        """
        Check the state of the subfiles.
        If the latest subfile is too large, create a new file and delete old ones.

        Parameters
        ----------
        potential_new_len: int, default 0
        """
        self.flush()

        latest_subfile_index = self.get_latest_subfile_index()
        latest_subfile_path = self.get_subfile_path_from_index(latest_subfile_index)

        ### First run with existing log files: open the most recent log file.
        is_first_run_with_logs = ((latest_subfile_index > -1) and self._current_file_obj is None)

        ### Sometimes a new file is created but output doesn't switch over.
        lost_latest_handle = (
            self._current_file_obj is not None
            and
            self.get_index_from_subfile_name(self._current_file_obj.name) == -1
        )
        if is_first_run_with_logs or lost_latest_handle:
            self._current_file_obj = open(latest_subfile_path, 'a+', encoding='utf-8')
            if self.redirect_streams:
                daemon.daemon.redirect_stream(sys.stdout, self._current_file_obj)
                daemon.daemon.redirect_stream(sys.stderr, self._current_file_obj)

        create_new_file = (
            (latest_subfile_index == -1)
            or
            self._current_file_obj is None
            or
            self.is_subfile_too_large(latest_subfile_index, potential_new_len)
        )
        if create_new_file:
            old_subfile_index = latest_subfile_index
            new_subfile_index = old_subfile_index + 1
            new_file_path = self.get_subfile_path_from_index(new_subfile_index)
            self._previous_file_obj = self._current_file_obj
            self._current_file_obj = open(new_file_path, 'a+', encoding='utf-8')
            self.subfile_objects[new_subfile_index] = self._current_file_obj
            self.flush()

            if self._previous_file_obj is not None:
                if self.redirect_streams:
                    self._redirected_subfile_objects[old_subfile_index] = self._previous_file_obj
                    daemon.daemon.redirect_stream(self._previous_file_obj, self._current_file_obj)
                    daemon.daemon.redirect_stream(sys.stdout, self._current_file_obj)
                    daemon.daemon.redirect_stream(sys.stderr, self._current_file_obj)
                self.close(unused_only=True)

            ### Sanity check in case writing somehow fails.
            if self._previous_file_obj is self._current_file_obj:
                self._previous_file_obj is None

            self.delete(unused_only=True)
        return self._current_file_obj


    def close(self, unused_only: bool = False) -> None:
        """
        Close any open file descriptors.

        Parameters
        ----------
        unused_only: bool, default False
            If `True`, only close file descriptors not currently in use.
        """
        subfile_indices = sorted(self.subfile_objects.keys())
        for subfile_index in subfile_indices:
            subfile_object = self.subfile_objects[subfile_index]
            if unused_only and subfile_object in (self._previous_file_obj, self._current_file_obj):
                continue
            try:
                if not subfile_object.closed:
                    #  subfile_object.flush()
                    subfile_object.close()
                _ = self.subfile_objects.pop(subfile_index, None)
                if self.redirect_streams:
                    _ = self._redirected_subfile_objects.pop(subfile_index, None)
            except Exception as e:
                warn(f"Failed to close an open subfile:\n{traceback.format_exc()}")

        if not unused_only:
            self._previous_file_obj = None
            self._current_file_obj = None


    def write(self, data: str) -> None:
        """
        Write the given text into the latest subfile.
        If the subfile will be too large, create a new subfile.
        If too many subfiles exist at once, the oldest one will be deleted.

        NOTE: This will not split data across multiple files.
        As such, if data is larger than max_file_size, then the corresponding subfile
        may exceed this limit.
        """
        self.file_path.parent.mkdir(exist_ok=True, parents=True)
        if isinstance(data, bytes):
            data = data.decode('utf-8')

        self.refresh_files(potential_new_len=len(data))
        try:
            self._current_file_obj.write(data)
        except Exception as e:
            warn(f"Failed to write to subfile:\n{traceback.format_exc()}")
        self.flush()
        self.delete(unused_only=True)


    def delete(self, unused_only: bool = False) -> None:
        """
        Delete old subfiles.

        Parameters
        ----------
        unused_only: bool, default False
            If `True`, only delete subfiles which are no longer needed.
        """
        existing_subfile_paths = self.get_existing_subfile_paths()
        if unused_only and len(existing_subfile_paths) <= self.num_files_to_keep:
            return

        self.flush()
        self.close(unused_only=unused_only)

        end_ix = (
            (-1 * self.num_files_to_keep)
            if unused_only
            else len(existing_subfile_paths)
        )
        for subfile_path_to_delete in existing_subfile_paths[0:end_ix]:
            subfile_index = self.get_index_from_subfile_name(subfile_path_to_delete.name)
            subfile_object = self.subfile_objects.get(subfile_index, None)

            try:
                subfile_path_to_delete.unlink()
            except Exception as e:
                warn(
                    f"Unable to delete subfile '{subfile_path_to_delete}':\n"
                    + f"{traceback.format_exc()}"
                )


    def read(self, *args, **kwargs) -> str:
        """
        Read the contents of the existing subfiles.
        """
        existing_subfile_indices = [
            self.get_index_from_subfile_name(subfile_path.name)
            for subfile_path in self.get_existing_subfile_paths()
        ]
        paths_to_read = [
            self.get_subfile_path_from_index(subfile_index)
            for subfile_index in existing_subfile_indices
            if subfile_index >= self._cursor[0]
        ]
        buffer = ''
        refresh_cursor = True
        for subfile_path in paths_to_read:
            subfile_index = self.get_index_from_subfile_name(subfile_path.name)
            seek_ix = (
                self._cursor[1]
                if subfile_index == self._cursor[0]
                else 0
            )

            if (
                subfile_index in self.subfile_objects
                and
                subfile_index not in self._redirected_subfile_objects
            ):
                subfile_object = self.subfile_objects[subfile_index]
                for i in range(self.SEEK_BACK_ATTEMPTS):
                    try:
                        subfile_object.seek(max(seek_ix - i, 0))
                        buffer += subfile_object.read()
                    except UnicodeDecodeError:
                        continue
                    break
            else:
                with open(subfile_path, 'r', encoding='utf-8') as f:
                    for i in range(self.SEEK_BACK_ATTEMPTS):
                        try:
                            f.seek(max(seek_ix - i, 0))
                            buffer += f.read()
                        except UnicodeDecodeError:
                            continue
                        break

                    ### Handle the case when no files have yet been opened.
                    if not self.subfile_objects and subfile_path == paths_to_read[-1]:
                        self._cursor = (subfile_index, f.tell())
                        refresh_cursor = False

        if refresh_cursor:
            self.refresh_cursor()
        return buffer


    def refresh_cursor(self) -> None:
        """
        Update the cursor to the latest subfile index and file.tell() value.
        """
        self.flush()
        existing_subfile_paths = self.get_existing_subfile_paths()
        current_ix = (
            self.get_index_from_subfile_name(existing_subfile_paths[-1].name)
            if existing_subfile_paths
            else 0
        )
        position = self._current_file_obj.tell() if self._current_file_obj is not None else 0
        self._cursor = (current_ix, position)


    def readlines(self) -> List[str]:
        """
        Return a list of lines of text.
        """
        existing_subfile_indices = [
            self.get_index_from_subfile_name(subfile_path.name)
            for subfile_path in self.get_existing_subfile_paths()
        ]
        paths_to_read = [
            self.get_subfile_path_from_index(subfile_index)
            for subfile_index in existing_subfile_indices
            if subfile_index >= self._cursor[0]
        ]

        lines = []
        refresh_cursor = True
        for subfile_path in paths_to_read:
            subfile_index = self.get_index_from_subfile_name(subfile_path.name)
            seek_ix = (
                self._cursor[1]
                if subfile_index == self._cursor[0]
                else 0
            )

            if (
                subfile_index in self.subfile_objects
                and
                subfile_index not in self._redirected_subfile_objects
            ):
                subfile_object = self.subfile_objects[subfile_index]
                for i in range(self.SEEK_BACK_ATTEMPTS):
                    try:
                        subfile_object.seek(max(seek_ix - i), 0)
                        subfile_lines = subfile_object.readlines()
                    except UnicodeDecodeError:
                        continue
                    break
            else:
                with open(subfile_path, 'r', encoding='utf-8') as f:
                    for i in range(self.SEEK_BACK_ATTEMPTS):
                        try:
                            f.seek(max(seek_ix - i, 0))
                            subfile_lines = f.readlines()
                        except UnicodeDecodeError:
                            continue
                        break

                    ### Handle the case when no files have yet been opened.
                    if not self.subfile_objects and subfile_path == paths_to_read[-1]:
                        self._cursor = (subfile_index, f.tell())
                        refresh_cursor = False

            ### Sometimes a line may span multiple files.
            if lines and subfile_lines and not lines[-1].endswith('\n'):
                lines[-1] += subfile_lines[0]
                new_lines = subfile_lines[1:]
            else:
                new_lines = subfile_lines
            lines.extend(new_lines)

        if refresh_cursor:
            self.refresh_cursor()
        return lines


    def seekable(self) -> bool:
        return True


    def seek(self, position: int) -> None:
        """
        Seek to the beginning of the logs stream.
        """
        existing_subfile_indices = self.get_existing_subfile_indices()
        min_ix = existing_subfile_indices[0] if existing_subfile_indices else 0
        max_ix = existing_subfile_indices[-1] if existing_subfile_indices else 0
        if position == 0:
            self._cursor = (min_ix, 0)
            return

        self._cursor = (max_ix, position)
        self._current_file_obj.seek(position)

    
    def flush(self) -> None:
        """
        Flush any open subfiles.
        """
        for subfile_index, subfile_object in self.subfile_objects.items():
            if not subfile_object.closed:
                try:
                    subfile_object.flush()
                except Exception as e:
                    warn(f"Failed to flush subfile:\n{traceback.format_exc()}")
        if self.redirect_streams:
            sys.stdout.flush()
            sys.stderr.flush()


    def __repr__(self) -> str:
        """
        Return basic info for this `RotatingFile`.
        """
        return (
            "RotatingFile("
            + f"'{self.file_path.as_posix()}', "
            + f"num_files_to_keep={self.num_files_to_keep}, "
            + f"max_file_size={self.max_file_size})"
        )



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Functions for editing elements belong here.
"""

from __future__ import annotations
import meerschaum as mrsm
from meerschaum.utils.typing import List, Any, SuccessTuple, Optional, Dict

def tag(
        action: Optional[List[str]] = None,
        **kwargs: Any
    ) -> SuccessTuple:
    """
    Edit an existing element.
    """
    from meerschaum.actions import choose_subaction
    options = {
        'pipes': _tag_pipes,
    }
    return choose_subaction(action, options, **kwargs)


def _tag_pipes(
        action: Optional[List[str]] = None,
        debug: bool = False,
        **kwargs: Any
    ) -> SuccessTuple:
    """
    Add or remove tags to registered pipes.
    Prefix a tag with a leading underscore to remove it.

    Note that the `--tags` flag applies to existing tags.
    Specify the tags you wish to add or remove as positional arguments, e.g.:

    ```
    mrsm tag pipes production sync-daily -c sql:main
    mrsm tag pipes _production --tags production
    ```
    """
    from meerschaum.utils.warnings import warn
    from meerschaum.utils.misc import separate_negation_values
    if not action:
        return False, "No tags were provided."

    pipes = mrsm.get_pipes(as_list=True, debug=debug, **kwargs)
    if not pipes:
        return False, "No pipes were found with the given keys."
    success_msg = f"Updated tags for {len(pipes)} pipe" + ('s' if len(pipes) != 1 else '') + '.'

    add_tags, remove_tags = separate_negation_values(action)
    edited_pipes = []
    for pipe in pipes:
        pipe_was_edited = False
        pipe_tags = pipe.tags
        existing_tags = set(pipe_tags)
        for tag_to_add in add_tags:
            if tag_to_add not in existing_tags:
                pipe_tags.append(tag_to_add)
                pipe_was_edited = True

        for tag_to_remove in remove_tags:
            if tag_to_remove in existing_tags:
                pipe_tags.remove(tag_to_remove)
                pipe_was_edited = True

        if pipe_was_edited:
            edited_pipes.append(pipe)

    if not edited_pipes:
        return True, success_msg

    failed_edit_msgs = []
    for pipe in edited_pipes:
        edit_success, edit_msg = pipe.edit(debug=debug)
        if not edit_success:
            warn(f"Failed to update tags for {pipe}:\n{edit_msg}", stack=False)
            failed_edit_msgs.append(edit_msg)

    num_failures = len(failed_edit_msgs)
    success = num_failures == 0
    msg = (
        success_msg
        if success else (
            f"Failed to update tags for {num_failures} pipe"
            + ('s' if num_failures != 1 else '')
            + ":\n"
            + '\n'.join(failed_edit_msgs)
        )
    )
    return success, msg


### NOTE: This must be the final statement of the module.
###       Any subactions added below these lines will not
###       be added to the `help` docstring.
from meerschaum.actions import choices_docstring as _choices_docstring
tag.__doc__ += _choices_docstring('tag')



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Open subprocesses and read their output
"""

from __future__ import annotations
from meerschaum.utils.typing import List, Any, SuccessTuple, Optional

def os(
        action: Optional[List[str]] = None,
        sub_args: Optional[List[str]] = None,
        debug: bool = False,
        **kw: Any
    ) -> SuccessTuple:
    """
    Launch a subprocess and read its output to stdout.
    """
    import os as _os
    from meerschaum.config import get_config

    import subprocess
    import sys
    from meerschaum.utils.debug import dprint

    if action is None:
        action = []

    ### determine system encoding
    encoding = sys.getdefaultencoding()

    command_list = []

    ### where to redirect stdout (default None)
    capture_stdout, capture_stderr = None, None
    timeout = None

    ### if actions are provided, don't launch interactive shell
    ### and read stdout, stderr and exit code
    if len(action) > 0:
        capture_stdout = subprocess.PIPE
        capture_stderr = subprocess.PIPE
        command_list += action + sub_args
        timeout = get_config('shell', 'timeout', patch=True)
    else:
        return False, os.__doc__

    if debug:
        dprint(command_list)

    process = subprocess.Popen(
        command_list,
        shell = False,
        stdout = capture_stdout,
        stderr = capture_stderr,
        env = _os.environ,
    )

    try:
        output_data, error_output_data = process.communicate(timeout=timeout)
    except Exception as e:
        print(e)
        process.kill()
        output_data, error_output_data = process.communicate()

    exit_code = process.wait()

    output = None
    if output_data is not None:
        output = output_data.decode(encoding)

    if output is not None:
        print(output, end="")
    else:
        output = "Success"

    error_output = "Error"
    if error_output_data is not None:
        error_output = error_output_data.decode(encoding)

    if debug:
        dprint("stdout:\n" + f"{output}")
        dprint("stderr:\n" + f"{error_output}")
        dprint(f"exit code: {exit_code}")

    if exit_code != 0:
        return (False, error_output)
    return (True, output)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Create a new SQLConnector from a URI.
"""

import os
from meerschaum.utils.typing import Optional, Dict, Any, Union
from meerschaum.utils.warnings import warn, error
from meerschaum.utils.packages import attempt_import

@classmethod
def from_uri(
        cls,
        uri: str,
        label: Optional[str] = None,
        as_dict: bool = False,
    ) -> Union[
        'meerschaum.connectors.SQLConnector',
        Dict[str, Union[str, int]],
    ]:
    """
    Create a new SQLConnector from a URI string.

    Parameters
    ----------
    uri: str
        The URI connection string.

    label: Optional[str], default None
        If provided, use this as the connector label.
        Otherwise use the determined database name.

    as_dict: bool, default False
        If `True`, return a dictionary of the keyword arguments
        necessary to create a new `SQLConnector`, otherwise create a new object.

    Returns
    -------
    A new SQLConnector object or a dictionary of attributes (if `as_dict` is `True`).
    """

    params = cls.parse_uri(uri)
    params['uri'] = uri
    flavor = params.get('flavor', None)
    if not flavor or flavor not in cls.flavor_configs:
        error(f"Invalid flavor '{flavor}' detected from the provided URI.")

    if 'database' not in params:
        error("Unable to determine the database from the provided URI.")

    if flavor in ('sqlite', 'duckdb'):
        if params['database'] == ':memory:':
            params['label'] = label or f'memory_{flavor}'
        else:
            params['label'] = label or params['database'].split(os.path.sep)[-1].lower()
    else:
        params['label'] = label or (
            (
                (params['username'] + '@' if 'username' in params else '')
                + params.get('host', '')
                + ('/' if 'host' in params else '')
                + params.get('database', '')
            ).lower()
        )

    return cls(**params) if not as_dict else params


@staticmethod
def parse_uri(uri: str) -> Dict[str, Any]:
    """
    Parse a URI string into a dictionary of parameters.

    Parameters
    ----------
    uri: str
        The database connection URI.

    Returns
    -------
    A dictionary of attributes.

    Examples
    --------
    >>> parse_uri('sqlite:////home/foo/bar.db')
    {'database': '/home/foo/bar.db', 'flavor': 'sqlite'}
    >>> parse_uri(
    ...     'mssql+pyodbc://sa:supersecureSECRETPASSWORD123!@localhost:1439'
    ...     + '/master?driver=ODBC+Driver+17+for+SQL+Server'
    ... )
    {'host': 'localhost', 'database': 'master', 'username': 'sa',
    'password': 'supersecureSECRETPASSWORD123!', 'port': 1439, 'flavor': 'mssql',
    'driver': 'ODBC Driver 17 for SQL Server'}
    >>> 
    """
    from urllib.parse import parse_qs, urlparse
    sqlalchemy = attempt_import('sqlalchemy')
    parser = sqlalchemy.engine.url.make_url
    params = parser(uri).translate_connect_args()
    params['flavor'] = uri.split(':')[0].split('+')[0]
    if params['flavor'] == 'postgres':
        params['flavor'] = 'postgresql'
    if '?' in uri:
        parsed_uri = urlparse(uri)
        for key, value in parse_qs(parsed_uri.query).items():
            params.update({key: value[0]})

        if '--search_path' in params.get('options', ''):
            params.update({'schema': params['options'].replace('--search_path=', '', 1)})
    return params



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

import pytest
import datetime
from decimal import Decimal
from tests import debug
from tests.connectors import conns, get_flavors
from tests.test_users import test_register_user
import meerschaum as mrsm
from meerschaum import Pipe
from meerschaum.actions import actions
from meerschaum.utils.dtypes import are_dtypes_equal
from meerschaum.utils.sql import sql_item_name

@pytest.fixture(autouse=True)
def run_before_and_after(flavor: str):
    test_register_user(flavor)
    yield

@pytest.mark.parametrize("flavor", get_flavors())
def test_sync_change_columns_dtypes(flavor: str):
    """
    Test that new columns are added.
    """
    conn = conns[flavor]
    pipe = Pipe('change', 'cols', 'dtypes', columns={'datetime': 'dt', 'id': 'id'}, instance=conn)
    pipe.delete(debug=debug)
    pipe = Pipe('change', 'cols', 'dtypes', columns={'datetime': 'dt', 'id': 'id'}, instance=conn)
    docs = [
        {'dt': '2022-01-01', 'id': 1, 'a': 10},
    ]
    pipe.sync(docs, debug=debug)
    assert len(pipe.get_data().columns) == 3

    docs = [
        {'dt': '2022-01-01', 'id': 1, 'a': 'foo'},
    ]
    pipe.sync(docs, debug=debug)
    df = pipe.get_data()
    assert len(df.columns) == 3
    assert len(df) == 1
    assert are_dtypes_equal(str(df.dtypes['a']), 'string')


@pytest.mark.parametrize("flavor", get_flavors())
def test_dtype_enforcement(flavor: str):
    """
    Test that incoming rows are enforced to the correct data type.
    """
    conn = conns[flavor]
    pipe = mrsm.Pipe('dtype', 'enforcement', instance=conn)
    pipe.delete(debug=debug)
    pipe = Pipe(
        'dtype', 'enforcement',
        columns = {
            'datetime': 'dt',
            'id': 'id',
        },
        dtypes = {
            'int': 'int',
            'float': 'float',
            'bool': 'bool',
            'object': 'object',
            'json': 'json',
            'numeric': 'numeric',
            'str': 'str',
        },
        instance = conn,
    )
    pipe.sync([{'dt': '2022-01-01', 'id': 1, 'int': '1'}], debug=debug)
    pipe.sync([{'dt': '2022-01-01', 'id': 1, 'float': '1.0'}], debug=debug)
    pipe.sync([{'dt': '2022-01-01', 'id': 1, 'bool': 'True'}], debug=debug)
    pipe.sync([{'dt': '2022-01-01', 'id': 1, 'object': 'foo'}], debug=debug)
    pipe.sync([{'dt': '2022-01-01', 'id': 1, 'str': 'bar'}], debug=debug)
    pipe.sync([{'dt': '2022-01-01', 'id': 1, 'json': '{"a": {"b": 1}}'}], debug=debug)
    pipe.sync([{'dt': '2022-01-01', 'id': 1, 'numeric': '1'}], debug=debug)
    df = pipe.get_data(debug=debug)
    assert len(df) == 1
    assert len(df.columns) == 9
    for col, typ in df.dtypes.items():
        pipe_dtype = pipe.dtypes[col]
        if pipe_dtype == 'json':
            assert isinstance(df[col][0], dict)
            pipe_dtype = 'object'
        elif pipe_dtype == 'numeric':
            assert isinstance(df[col][0], Decimal)
            pipe_dtype = 'object'
        elif pipe_dtype == 'str':
            assert isinstance(df[col][0], str)
        assert are_dtypes_equal(pipe_dtype.lower(), typ)


@pytest.mark.parametrize("flavor", get_flavors())
def test_infer_json_dtype(flavor: str):
    """
    Ensure that new pipes with complex columns (dict or list) as enforced as JSON. 
    """
    from meerschaum.utils.formatting import pprint
    from meerschaum.utils.misc import generate_password
    session_id = generate_password(6)
    conn = conns[flavor]
    pipe = Pipe('foo', 'bar', session_id, instance=conn)
    _ = pipe.delete(debug=debug)
    pipe = Pipe('foo', 'bar', session_id, instance=conn)
    success, msg = pipe.sync([
        {'a': ['b', 'c']},
        {'a': {'b': 1}},
    ])
    assert success, msg
    pprint(pipe.get_columns_types())
    df = pipe.get_data(debug=debug)
    print(df)
    assert isinstance(df['a'][0], list)
    assert isinstance(df['a'][1], dict)


@pytest.mark.parametrize("flavor", get_flavors())
def test_force_json_dtype(flavor: str):
    """
    Ensure that new pipes with complex columns (dict or list) as enforced as JSON. 
    """
    import json
    from meerschaum.utils.formatting import pprint
    from meerschaum.utils.misc import generate_password
    session_id = generate_password(6)
    conn = conns[flavor]
    pipe = Pipe('foo', 'bar', session_id, instance=conn, dtypes={'a': 'json'})
    _ = pipe.delete(debug=debug)
    pipe = Pipe('foo', 'bar', session_id, instance=conn, dtypes={'a': 'json'})
    success, msg = pipe.sync([
        {'a': json.dumps(['b', 'c'])},
        {'a': json.dumps({'b': 1})},
    ])
    assert success, msg
    pprint(pipe.get_columns_types())
    df = pipe.get_data(debug=debug)
    print(df)
    assert isinstance(df['a'][0], list)
    assert isinstance(df['a'][1], dict)


@pytest.mark.parametrize("flavor", get_flavors())
def test_infer_numeric_dtype(flavor: str):
    """
    Ensure that `Decimal` objects are persisted as `numeric`. 
    """
    from meerschaum.utils.formatting import pprint
    from meerschaum.utils.misc import generate_password
    from meerschaum.utils.dtypes.sql import NUMERIC_PRECISION_FLAVORS
    scale, precision = NUMERIC_PRECISION_FLAVORS.get(
        flavor, NUMERIC_PRECISION_FLAVORS['sqlite']
    )
    digits = (list(reversed(range(0, 10))) * 4)[:(-1 * (40 - scale))]
    decimal_digits = digits[(-1 * precision):]
    numeric_digits = digits[:(-1 * precision)]
    numeric_str = (
        ''.join([str(digit) for digit in numeric_digits])
        + '.' +
        ''.join([str(digit) for digit in decimal_digits])
    )
    conn = conns[flavor]
    pipe = Pipe('infer', 'numeric', instance=conn)
    _ = pipe.delete(debug=debug)
    pipe = Pipe('infer', 'numeric', instance=conn)
    success, msg = pipe.sync([
        {'a': Decimal('1')},
        {'a': Decimal(numeric_str)},
    ])
    assert success, msg
    pprint(pipe.get_columns_types())
    df = pipe.get_data(debug=debug)
    print(df)
    assert isinstance(df['a'][0], Decimal)
    assert df['a'][0] == Decimal('1')
    assert isinstance(df['a'][1], Decimal)
    assert df['a'][1] == Decimal(numeric_str)


@pytest.mark.parametrize("flavor", get_flavors())
def test_infer_numeric_from_mixed_types(flavor: str):
    """
    Ensure that new pipes with complex columns (dict or list) as enforced as NUMERIC. 
    """
    from meerschaum.utils.formatting import pprint
    from meerschaum.utils.misc import generate_password
    session_id = generate_password(6)
    conn = conns[flavor]
    pipe = Pipe('foo', 'bar', session_id, instance=conn)
    _ = pipe.delete(debug=debug)
    pipe = Pipe('foo', 'bar', session_id, instance=conn)
    success, msg = pipe.sync([{'a': 1}])
    assert success, msg
    success, msg = pipe.sync([{'a': 2.1}])
    assert success, msg
    pprint(pipe.get_columns_types())
    df = pipe.get_data(debug=debug)
    print(df)
    assert isinstance(df['a'][0], Decimal)
    assert df['a'][0] == Decimal('1')
    assert isinstance(df['a'][1], Decimal)
    assert df['a'][1] == Decimal('2.1')


@pytest.mark.parametrize("flavor", get_flavors())
def test_force_numeric_dtype(flavor: str):
    """
    Ensure that new pipes with complex columns (dict or list) as enforced as NUMERIC. 
    """
    from meerschaum.utils.formatting import pprint
    from meerschaum.utils.misc import generate_password
    session_id = generate_password(6)
    conn = conns[flavor]
    pipe = Pipe('foo', 'bar', session_id, instance=conn, dtypes={'a': 'numeric'})
    _ = pipe.delete(debug=debug)
    pipe = Pipe('foo', 'bar', session_id, instance=conn, dtypes={'a': 'numeric'})
    success, msg = pipe.sync([
        {'a': '1'},
        {'a': '2.1'},
    ])
    assert success, msg
    pprint(pipe.get_columns_types())
    df = pipe.get_data(debug=debug)
    print(df)
    assert isinstance(df['a'][0], Decimal)
    assert df['a'][0] == Decimal('1')
    assert isinstance(df['a'][1], Decimal)
    assert df['a'][1] == Decimal('2.1')

    success, msg = pipe.sync([{'a': None}], debug=debug)
    assert success, msg
    pprint(pipe.get_columns_types())
    df = pipe.get_data(debug=debug)
    docs = df.to_dict(orient='records')
    assert str(docs[-1]['a']) == 'NaN'
    for doc in docs:
        assert isinstance(doc['a'], Decimal)


@pytest.mark.parametrize("flavor", get_flavors())
def test_utc_offset_datetimes(flavor: str):
    """
    Verify that we are able to sync rows with UTC offset datetimes.
    """
    conn = conns[flavor]
    pipe = Pipe(
        'test_utc_offset', 'datetimes',
        instance = conn,
        columns = {'datetime': 'dt'},
    )
    pipe.delete()

    docs = [
        {'dt': '2023-01-01 00:00:00+00:00'},
        {'dt': '2023-01-02 00:00:00+01:00'},
    ]

    expected_docs = [
        {'dt': datetime.datetime(2023, 1, 1)},
        {'dt': datetime.datetime(2023, 1, 1, 23, 0, 0)}
    ]

    success, msg = pipe.sync(docs, debug=debug)
    assert success, msg
    df = pipe.get_data(debug=debug)
    synced_docs = df.to_dict(orient='records')
    assert synced_docs == expected_docs


@pytest.mark.parametrize("flavor", get_flavors())
def test_ignore_datetime_conversion(flavor: str):
    """
    If the user specifies, skip columns from being detected as datetimes.
    """
    conn = conns[flavor]
    pipe = Pipe(
        'test_utc_offset', 'datetimes', 'ignore',
        instance = conn,
        dtypes = {
            'dt': 'str',
        },
    )
    pipe.delete()

    docs = [
        {'dt': '2023-01-01 00:00:00+00:00'},
        {'dt': '2023-01-02 00:00:00+01:00'},
    ]

    expected_docs = [
        {'dt': '2023-01-01 00:00:00+00:00'},
        {'dt': '2023-01-02 00:00:00+01:00'},
    ]

    success, msg = pipe.sync(docs, debug=debug)
    assert success, msg
    df = pipe.get_data(debug=debug)
    synced_docs = df.to_dict(orient='records')
    assert synced_docs == expected_docs


@pytest.mark.parametrize("flavor", get_flavors())
def test_no_indices_inferred_datetime_to_text(flavor: str):
    """
    Verify that changing dtypes are handled.
    """
    conn = conns[flavor]
    pipe = Pipe(
        'test_no_indices', 'datetimes', 'text',
        instance = conn,
    )
    pipe.delete()
    docs = [
        {'fake-dt': '2023-01-01', 'a': 1},
    ]
    success, msg = pipe.sync(docs, debug=debug)
    assert success, msg

    docs = [
        {'fake-dt': '2023-01-01', 'a': 1},
        {'fake-dt': '2023-01-02', 'a': 2},
    ]
    success, msg = pipe.sync(docs, debug=debug)
    assert success, msg
    df = pipe.get_data()
    assert len(df) == len(docs)

    docs = [
        {'fake-dt': '2023-01-01', 'a': 1},
        {'fake-dt': '2023-01-02', 'a': 2},
        {'fake-dt': 'foo', 'a': 3},
    ]
    success, msg = pipe.sync(docs, debug=debug)
    assert success, msg
    df = pipe.get_data()
    assert len(df) == len(docs)


@pytest.mark.parametrize("flavor", get_flavors())
def test_sync_bools(flavor: str):
    """
    Test that pipes are able to sync bools.
    """
    conn = conns[flavor]
    pipe = mrsm.Pipe('test', 'bools', instance=conn)
    _ = pipe.delete()
    pipe = mrsm.Pipe(
        'test', 'bools',
        instance = conn,
        columns = {'datetime': 'dt'},
        dtypes = {'is_bool': 'bool'},
    )
    _ = pipe.drop()
    docs = [
        {'dt': '2023-01-01', 'is_bool': True},
        {'dt': '2023-01-02', 'is_bool': False},
    ]
    success, msg = pipe.sync(docs, debug=False)
    assert success, msg

    df = pipe.get_data()
    assert 'bool' in str(df.dtypes['is_bool'])

    synced_docs = [
        {
            'dt': doc['dt'].strftime('%Y-%m-%d'),
            'is_bool': doc['is_bool'],
        }
        for doc in df.to_dict(orient='records')
    ]
    assert synced_docs == docs

    new_docs = [
        {'dt': '2023-01-01', 'is_bool': False},
        {'dt': '2023-01-02', 'is_bool': True},
    ]
    success, msg = pipe.sync(new_docs, debug=debug)
    assert success, msg

    df = pipe.get_data()
    synced_docs = [
        {
            'dt': doc['dt'].strftime('%Y-%m-%d'),
            'is_bool': doc['is_bool'],
        }
        for doc in df.to_dict(orient='records')
    ]

    assert synced_docs == new_docs


@pytest.mark.parametrize("flavor", get_flavors())
def test_sync_bools_inferred(flavor: str):
    """
    Test that pipes are able to sync bools.
    """
    conn = conns[flavor]
    pipe = mrsm.Pipe('test', 'bools', instance=conn)
    _ = pipe.delete()
    pipe = mrsm.Pipe(
        'test', 'bools',
        instance = conn,
        columns = {'datetime': 'dt'},
    )
    _ = pipe.drop()
    docs = [
        {'dt': '2023-01-01', 'is_bool': True},
        {'dt': '2023-01-02', 'is_bool': False},
    ]
    success, msg = pipe.sync(docs, debug=True)
    assert success, msg

    df = pipe.get_data()
    assert 'bool' in str(df.dtypes['is_bool'])

    synced_docs = [
        {
            'dt': doc['dt'].strftime('%Y-%m-%d'),
            'is_bool': doc['is_bool'],
        }
        for doc in df.to_dict(orient='records')
    ]
    assert synced_docs == docs

    new_docs = [
        {'dt': '2023-01-01', 'is_bool': False},
        {'dt': '2023-01-02', 'is_bool': True},
    ]
    success, msg = pipe.sync(new_docs, debug=debug)
    assert success, msg

    df = pipe.get_data()
    synced_docs = [
        {
            'dt': doc['dt'].strftime('%Y-%m-%d'),
            'is_bool': doc['is_bool'],
        }
        for doc in df.to_dict(orient='records')
    ]

    assert synced_docs == new_docs


@pytest.mark.parametrize("flavor", get_flavors())
def test_sync_bools_inplace(flavor: str):
    """
    Test that pipes are able to sync bool in-place.
    """
    conn = conns[flavor]
    pipe = mrsm.Pipe('test', 'bools', 'inplace', instance=conn)
    _ = pipe.delete()
    pipe = mrsm.Pipe(
        'test', 'bools', 'inplace',
        instance = conn,
        columns = {'datetime': 'dt', 'id': 'id'},
    )
    pipe_table = sql_item_name(pipe.target, conn.flavor) if conn.type == 'sql' else pipe.target
    inplace_pipe = mrsm.Pipe(conn, 'bools', 'inplace', instance=conn)
    _ = inplace_pipe.delete()
    inplace_pipe = mrsm.Pipe(
        conn, 'bools', 'inplace',
        instance = conn,
        columns = pipe.columns,
        dtypes = {
            'is_bool': 'bool',
        },
        parameters = {
            'fetch': {
                'definition': f"SELECT * FROM {pipe_table}",
                'pipe': pipe.keys(),
            },
        },
    )
    _ = pipe.drop()
    docs = [
        {'dt': '2023-01-01', 'id': 1, 'is_bool': True},
        {'dt': '2023-01-02', 'id': 2, 'is_bool': False},
        {'dt': '2023-01-03', 'id': 3, 'is_bool': None},
    ]
    success, msg = pipe.sync(docs)
    assert success, msg

    success, msg = inplace_pipe.sync(debug=debug)
    assert success, msg

    assert 'bool' in inplace_pipe.dtypes['is_bool']
    assert inplace_pipe.get_rowcount() == len(docs)

    df = inplace_pipe.get_data()

    pipe.sync([{'dt': '2023-01-03', 'id': 3, 'is_bool': True}])
    success, msg = inplace_pipe.sync(debug=debug)
    assert success, msg
    df = inplace_pipe.get_data(params={'id': 3})
    assert 'true' in str(df['is_bool'][0]).lower()

    pipe.sync([{'dt': '2023-01-03', 'id': 3, 'is_bool': None}])
    success, msg = inplace_pipe.sync(debug=debug)
    assert success, msg
    df = inplace_pipe.get_data(params={'id': 3})
    assert 'na' in str(df['is_bool'][0]).lower()




#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Define a context manager for virtual environments.
"""

from __future__ import annotations

import copy
import pathlib
from meerschaum.utils.typing import Union
from meerschaum.utils.threading import RLock


class Venv:
    """
    Manage a virtual enviroment's activation status.

    Examples
    --------
    >>> from meerschaum.plugins import Plugin
    >>> with Venv('mrsm') as venv:
    ...     import pandas
    >>> with Venv(Plugin('noaa')) as venv:
    ...     import requests
    >>> venv = Venv('mrsm')
    >>> venv.activate()
    True
    >>> venv.deactivate()
    True
    >>> 
    """

    def __init__(
            self,
            venv: Union[str, 'meerschaum.plugins.Plugin', None] = 'mrsm',
            debug: bool = False,
        ) -> None:
        from meerschaum.utils.venv import activate_venv, deactivate_venv, active_venvs
        ### For some weird threading issue,
        ### we can't use `isinstance` here.
        if 'meerschaum.plugins._Plugin' in str(type(venv)):
            self._venv = venv.name
            self._activate = venv.activate_venv
            self._deactivate = venv.deactivate_venv
            self._kwargs = {}
        else:
            self._venv = venv
            self._activate = activate_venv
            self._deactivate = deactivate_venv
            self._kwargs = {'venv': venv}
        self._debug = debug
        ### In case someone calls `deactivate()` before `activate()`.
        self._kwargs['previously_active_venvs'] = copy.deepcopy(active_venvs)


    def activate(self, debug: bool = False) -> bool:
        """
        Activate this virtual environment.
        If a `meerschaum.plugins.Plugin` was provided, its dependent virtual environments
        will also be activated.
        """
        from meerschaum.utils.venv import active_venvs
        self._kwargs['previously_active_venvs'] = copy.deepcopy(active_venvs)
        return self._activate(debug=(debug or self._debug), **self._kwargs)


    def deactivate(self, debug: bool = False) -> bool:
        """
        Deactivate this virtual environment.
        If a `meerschaum.plugins.Plugin` was provided, its dependent virtual environments
        will also be deactivated.
        """
        return self._deactivate(debug=(debug or self._debug), **self._kwargs)


    @property
    def target_path(self) -> pathlib.Path:
        """
        Return the target site-packages path for this virtual environment.
        A `meerschaum.utils.venv.Venv` may have one virtual environment per minor Python version
        (e.g. Python 3.10 and Python 3.7).
        """
        from meerschaum.utils.venv import venv_target_path
        return venv_target_path(venv=self._venv, allow_nonexistent=True, debug=self._debug)


    @property
    def root_path(self) -> pathlib.Path:
        """
        Return the top-level path for this virtual environment.
        """
        from meerschaum.config._paths import VIRTENV_RESOURCES_PATH
        if self._venv is None:
            return self.target_path.parent
        return VIRTENV_RESOURCES_PATH / self._venv


    def __enter__(self) -> None:
        self.activate(debug=self._debug)


    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        self.deactivate(debug=self._debug)


    def __str__(self) -> str:
        quote = "'" if self._venv is not None else ""
        return "Venv(" + quote + str(self._venv) + quote + ")"


    def __repr__(self) -> str:
        return self.__str__()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Synchronize pipes' data with sources.

NOTE: `sync` required a SQL connection and is not intended for client use
"""

from __future__ import annotations
from datetime import timedelta, datetime, timezone
import meerschaum as mrsm
from meerschaum.utils.typing import SuccessTuple, Any, List, Optional, Tuple, Union

def sync(
        action: Optional[List[str]] = None,
        **kw: Any
    ) -> SuccessTuple:
    """
    Fetch and sync data for pipes.

    Usage:
        - `--loop`
            - Sync indefinitely.
        - `--min-seconds 10`
            - Wait 10 seconds between laps.
        - `--async`, `--unblock``
            - Spin up background threads for each pipe.
        - `--debug`
            - Print verbose messages.
    """
    from meerschaum.actions import choose_subaction
    options = {
        'pipes': _sync_pipes,
    }
    return choose_subaction(action, options, **kw)


def _pipes_lap(
        workers: Optional[int] = None,
        debug: Optional[bool] = None,
        unblock: bool = False,
        force: bool = False,
        min_seconds: int = 1,
        verify: bool = False,
        deduplicate: bool = False,
        bounded: Optional[bool] = None,
        chunk_interval: Union[timedelta, int, None] = None,
        mrsm_instance: Optional[str] = None,
        timeout_seconds: Optional[int] = None,
        nopretty: bool = False,
        _progress: Optional['rich.progress.Progress'] = None,
        **kw: Any
    ) -> Tuple[List[mrsm.Pipe], List[mrsm.Pipe]]:
    """
    Do a lap of syncing pipes.
    """
    from meerschaum import get_pipes
    from meerschaum.utils.debug import dprint, _checkpoint
    from meerschaum.utils.packages import attempt_import, import_rich
    from meerschaum.utils.formatting import print_tuple, ANSI, UNICODE, get_console
    from meerschaum.utils.warnings import warn
    from meerschaum.utils.threading import Lock, RLock, Thread, Event
    from meerschaum.utils.misc import print_options, get_cols_lines
    from meerschaum.utils.pool import get_pool_executor, get_pool
    from meerschaum.connectors.parse import parse_instance_keys
    from meerschaum import Plugin
    import queue
    import multiprocessing
    import contextlib
    import time, os, copy
    from meerschaum.utils.packages import venv_exec
    from meerschaum.utils.process import poll_process
    import json
    import sys
    dill = attempt_import('dill')

    rich_table, rich_text, rich_box = attempt_import(
        'rich.table', 'rich.text', 'rich.box',
        lazy = False,
    )
    all_kw = copy.deepcopy(kw)
    all_kw.update({
        'workers': workers,
        'debug': debug,
        'unblock': unblock,
        'force': force,
        'min_seconds': min_seconds,
        'timeout_seconds': timeout_seconds,
        'mrsm_instance': mrsm_instance,
        'verify': verify,
        'deduplicate': deduplicate,
        'bounded': bounded,
        'chunk_interval': chunk_interval,
    })
    locks = {'remaining_count': Lock(), 'results_dict': Lock(), 'pipes_threads': Lock(),}
    pipes = get_pipes(
        as_list=True, method='registered', debug=debug, mrsm_instance=mrsm_instance, **kw
    )
    remaining_count = len(pipes)
    instance_connector = parse_instance_keys(mrsm_instance, debug=debug)
    conns = (
        instance_connector.engine.pool.size() if instance_connector.type == 'sql'
        else len(pipes)
    )
    cores = multiprocessing.cpu_count()
    pipes_queue = queue.Queue(remaining_count)
    for pipe in pipes:
        pipes_queue.put_nowait(pipe)
    stop_event = Event()
    results_dict = {}

    ### Cap the number workers to the pool size or 1 if working in-memory.
    if workers is None:
        workers = (
            1 if (
                instance_connector.type == 'sql'
                and
                instance_connector.__dict__.get('database', None) == ':memory:'
            ) else min(cores, (conns if conns != 0 else cores))
        )
    if workers > conns and conns != 0 and instance_connector.type == 'sql':
        warn(
            f"Using more workers ({workers}) than the available pool of database connections "
            + "may lead to concurrency issues.\n    You can change the pool size with "
            + "`edit config system` under the keys connectors:sql:create_engine:pool_size,\n    "
            + "and a size of 0 will not limit the number of connections.",
            stack = False,
        )


    def _task_label(count: int):
        return f"[cyan]Syncing {count} pipe{'s' if count != 1 else ''}..."

    _task = (
        _progress.add_task(_task_label(len(pipes)), start=True, total=len(pipes))
    ) if _progress is not None else None


    def worker_fn():
        while not stop_event.is_set():
            try:
                pipe = pipes_queue.get_nowait()
            except queue.Empty:
                return
            return_tuple = sync_pipe(pipe)
            results_dict[pipe] = return_tuple

            if not nopretty:
                success, msg = return_tuple
                msg = (
                    f"Finished syncing {pipe}:\n" if success
                    else f"Error while syncing {pipe}:\n"
                ) + msg + '\n'
                print_tuple(
                    (success, msg),
                    calm = True,
                    _progress = _progress,
                )
            _checkpoint(_progress=_progress, _task=_task)
            if _progress is not None:
                nonlocal remaining_count
                with locks['remaining_count']:
                    remaining_count -= 1
            pipes_queue.task_done()

    sync_function_source = dill.source.getsource(_wrap_pipe)
    fence_begin, fence_end = '<MRSM_RESULT>', '</MRSM_RESULT>'

    def sync_pipe(p):
        """
        Wrapper function for handling exceptions.
        """
        ### If no timeout is specified, handle syncing in the current thread.
        if timeout_seconds is None:
            return _wrap_pipe(p, **all_kw)
        _success_tuple = False, "Nothing returned."
        def write_line(line):
            nonlocal _success_tuple
            decoded = line.decode('utf-8')
            begin_index, end_index = decoded.find(fence_begin), decoded.find(fence_end)

            ### Found the beginning of the return value.
            ### Don't write the parsed success tuple message.
            if begin_index >= 0:
                _success_tuple = tuple(json.loads(
                    decoded[begin_index + len(fence_begin):end_index]
                ))
                return
            sys.stdout.buffer.write(line)

        def timeout_handler(p, *args, **kw):
            success, msg = False, (
                f"Failed to sync {p} within {timeout_seconds} second"
                + ('s' if timeout_seconds != 1 else '') + '.'
            )
            write_line((fence_begin + json.dumps((success, msg)) + fence_end).encode('utf-8'))

        src = (
            "from __future__ import annotations\n"
            + "from datetime import timedelta\n"
            + "from typing import Optional, Union\n"
            + "\n\n"
            + sync_function_source
            + '\n\n'
            + "import meerschaum as mrsm\n"
            + "import json\n"
            + f"pipe = mrsm.Pipe(**json.loads({json.dumps(json.dumps(p.meta))}))\n"
            + f"""print(
                    '{fence_begin}'
                    + json.dumps(
                        _wrap_pipe(
                            pipe,
                            **json.loads({json.dumps(json.dumps(all_kw))})
                        )
                    )
                    + '{fence_end}'
            )"""
        )
        if debug:
            dprint(src)

        proc = venv_exec(src, venv=None, as_proc=True, debug=debug)
        poll_process(
            proc,
            write_line,
            timeout_seconds,
            timeout_handler,
            (p,)
        )
        return _success_tuple

    worker_threads = [Thread(target=worker_fn) for _ in range(min(workers, len(pipes)))]
    for worker_thread in worker_threads:
        worker_thread.start()

    try:
        while any([t.is_alive() for t in worker_threads]):
            time.sleep(0.1)
    except KeyboardInterrupt:
        stop_event.set()
        raise

    pipes_queue.join()
    for worker_thread in worker_threads:
        worker_thread.join()
    return results_dict


def _sync_pipes(
        loop: bool = False,
        min_seconds: int = 1,
        unblock: bool = False,
        verify: bool = False,
        deduplicate: bool = False,
        bounded: Optional[bool] = None,
        chunk_interval: Union[timedelta, int, None] = None,
        shell: bool = False,
        nopretty: bool = False,
        debug: bool = False,
        **kw: Any
    ) -> SuccessTuple:
    """
    Fetch and sync new data for pipes.

    Usage:
        - `--loop`
            - Sync indefinitely.
        - `--min-seconds 10`
            - Wait 10 seconds between laps.
        - `--async`, `--unblock``
            - Spin up background threads for each pipe.
        - `--debug`
            - Print verbose messages.
    """
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.warnings import warn, info
    from meerschaum.utils.formatting import UNICODE
    from meerschaum.utils.formatting._shell import progress, live
    from meerschaum.utils.formatting._shell import clear_screen, flush_with_newlines
    from meerschaum.utils.formatting import print_pipes_results
    import contextlib
    import time
    import sys
    import json
    import asyncio
    run = True
    msg = ""
    interrupt_warning_msg = "Syncing was interrupted due to a keyboard interrupt."
    cooldown = 2 * (min_seconds + 1)
    underline = '\u2015' if UNICODE else '-'
    success_pipes, failure_pipes = [], []
    while run:
        _progress = progress() if shell else None
        cm = _progress if _progress is not None else contextlib.nullcontext()

        lap_begin = time.perf_counter()

        try:
            results_dict = {}
            with cm:
                results_dict = _pipes_lap(
                    min_seconds = min_seconds,
                    _progress = _progress,
                    verify = verify,
                    deduplicate = deduplicate,
                    bounded = bounded,
                    chunk_interval = chunk_interval,
                    unblock = unblock,
                    debug = debug,
                    nopretty = nopretty,
                    **kw
                )
                success_pipes = [
                    pipe
                    for pipe, (_success, _msg) in results_dict.items()
                    if _success
                ]
                failure_pipes = [
                    pipe
                    for pipe, (_success, _msg) in results_dict.items()
                    if not _success
                ]
        except Exception as e:
            import traceback
            traceback.print_exc()
            warn(
                f"Failed to sync all pipes. Waiting for {cooldown} seconds, then trying again.",
                stack = False
            )
            results_dict = {}
            success_pipes, failure_pipes = None, None
            try:
                time.sleep(cooldown)
            except KeyboardInterrupt:
                warn(interrupt_warning_msg, stack=False)
                loop, run = False, False
            else:
                cooldown = int(cooldown * 1.5)
                continue
        except KeyboardInterrupt:
            warn(interrupt_warning_msg, stack=False)
            loop, run = False, False
        cooldown = 2 * (min_seconds + 1)
        lap_end = time.perf_counter()
        print()

        if success_pipes is not None and not loop and shell and not nopretty:
            clear_screen(debug=debug)

        success_msg = (
            "Successfully spawned threads for pipes:"
            if unblock
            else f"Successfully synced pipes:"
        )
        fail_msg = f"Failed to sync pipes:"
        if results_dict:
            print_pipes_results(
                results_dict,
                success_header = success_msg,
                failure_header = fail_msg,
                nopretty = nopretty,
            )

        msg = (
            f"It took {round(lap_end - lap_begin, 2)} seconds to sync " +
            f"{len(success_pipes) + len(failure_pipes)} pipe" +
                ("s" if (len(success_pipes) + len(failure_pipes)) != 1 else "") + "\n" +
            f"    ({len(success_pipes)} succeeded, {len(failure_pipes)} failed)."
        ) if success_pipes is not None else "Syncing was aborted."
        if min_seconds > 0 and loop:
            print()
            info(
                f"Sleeping for {min_seconds} second" +
                ("s" if abs(min_seconds) != 1 else "")
                + '.'
            )
            try:
                time.sleep(min_seconds)
            except KeyboardInterrupt:
                loop, run = False, False
                warn(interrupt_warning_msg, stack=False)
        run = loop
    return (len(success_pipes) > 0 if success_pipes is not None else False), msg


def _wrap_pipe(
        pipe,
        unblock: bool = False,
        force: bool = False,
        debug: bool = False,
        min_seconds: int = 1,
        workers = None,
        verify: bool = False,
        deduplicate: bool = False,
        bounded: Optional[bool] = None,
        chunk_interval: Union[timedelta, int, None] = None,
        **kw
    ):
    """
    Wrapper function for handling exceptions.
    """
    import time
    import traceback
    from datetime import datetime, timedelta, timezone
    import meerschaum as mrsm
    from meerschaum.utils.typing import is_success_tuple, SuccessTuple
    from meerschaum.connectors import get_connector_plugin
    from meerschaum.utils.venv import Venv
    from meerschaum.plugins import _pre_sync_hooks, _post_sync_hooks
    from meerschaum.utils.misc import filter_keywords
    from meerschaum.utils.pool import get_pool
    from meerschaum.utils.warnings import warn

    pool = get_pool(workers=workers)

    sync_timestamp = datetime.now(timezone.utc)
    sync_start = time.perf_counter()
    sync_kwargs = {k: v for k, v in kw.items() if k != 'blocking'}
    sync_kwargs.update({
        'blocking': (not unblock),
        'force': force,
        'debug': debug,
        'min_seconds': min_seconds,
        'workers': workers,
        'bounded': bounded,
        'chunk_interval': chunk_interval,
        'sync_timestamp': sync_timestamp,
    })
    if not verify and not deduplicate:
        sync_method = pipe.sync
    elif not verify and deduplicate:
        sync_method = pipe.deduplicate
    else:
        sync_method = pipe.verify
        sync_kwargs['deduplicate'] = deduplicate
    sync_kwargs['sync_method'] = sync_method

    def call_sync_hook(plugin_name: str, sync_hook) -> SuccessTuple:
        plugin = mrsm.Plugin(plugin_name) if plugin_name else None
        with mrsm.Venv(plugin):
            try:
                sync_hook_result = sync_hook(pipe, **filter_keywords(sync_hook, **sync_kwargs))
                if is_success_tuple(sync_hook_result):
                    return sync_hook_result
            except Exception as e:
                msg = (
                    f"Failed to execute sync hook '{sync_hook.__name__}' "
                    + f"from plugin '{plugin}':\n{traceback.format_exc()}"
                )
                warn(msg, stack=False)
                return False, msg
        return True, "Success"

    pre_hook_results, post_hook_results = [], []
    def apply_hooks(is_pre_sync: bool):
        _sync_hooks = (_pre_sync_hooks if is_pre_sync else _post_sync_hooks)
        _hook_results = (pre_hook_results if is_pre_sync else post_hook_results)
        for module_name, sync_hooks in _sync_hooks.items():
            plugin_name = module_name.split('.')[-1] if module_name.startswith('plugins.') else None
            for sync_hook in sync_hooks:
                hook_result = pool.apply_async(call_sync_hook, (plugin_name, sync_hook))
                _hook_results.append(hook_result)

    apply_hooks(True)
    for hook_result in pre_hook_results:
        hook_success, hook_msg = hook_result.get()
        mrsm.pprint((hook_success, hook_msg))
   
    try:
        with Venv(get_connector_plugin(pipe.connector), debug=debug):
            return_tuple = sync_method(**sync_kwargs)
    except Exception as e:
        import traceback
        traceback.print_exception(type(e), e, e.__traceback__)
        print("Error: " + str(e))
        return_tuple = (False, f"Failed to sync {pipe} with exception:" + "\n" + str(e))

    duration = time.perf_counter() - sync_start
    sync_kwargs.update({
        'success_tuple': return_tuple,
        'sync_duration': duration,
        'sync_complete_timestamp': datetime.now(timezone.utc),
    })
    apply_hooks(False)
    for hook_result in post_hook_results:
        hook_success, hook_msg = hook_result.get()
        mrsm.pprint((hook_success, hook_msg))

    return return_tuple



### NOTE: This must be the final statement of the module.
###       Any subactions added below these lines will not
###       be added to the `help` docstring.
from meerschaum.actions import choices_docstring as _choices_docstring
sync.__doc__ += _choices_docstring('sync')



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Drop a Pipe's table but keep its registration
"""

from __future__ import annotations
from meerschaum.utils.typing import SuccessTuple, Any

def drop(
        self,
        debug: bool = False,
        **kw: Any
    ) -> SuccessTuple:
    """
    Call the Pipe's instance connector's `drop_pipe()` method.

    Parameters
    ----------
    debug: bool, default False:
        Verbosity toggle.

    Returns
    -------
    A `SuccessTuple` of success, message.

    """
    self._exists = False
    from meerschaum.utils.warnings import warn
    from meerschaum.utils.venv import Venv
    from meerschaum.connectors import get_connector_plugin

    if self.cache_pipe is not None:
        _drop_cache_tuple = self.cache_pipe.drop(debug=debug, **kw)
        if not _drop_cache_tuple[0]:
            warn(_drop_cache_tuple[1])

    with Venv(get_connector_plugin(self.instance_connector)):
        result = self.instance_connector.drop_pipe(self, debug=debug, **kw)
    return result



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
This module is the entry point for the interactive shell.
"""

from __future__ import annotations
import os
from copy import deepcopy
from meerschaum.utils.typing import Union, SuccessTuple, Any, Callable, Optional, List, Dict
from meerschaum.utils.packages import attempt_import
from meerschaum.config import __doc__, __version__ as version, get_config
import cmd
_old_input = cmd.__builtins__['input']
prompt_toolkit = attempt_import('prompt_toolkit', lazy=False, warn=False, install=True)
(
    prompt_toolkit_shortcuts,
    prompt_toolkit_history,
    prompt_toolkit_formatted_text,
    prompt_toolkit_styles,
) = (
    attempt_import('prompt_toolkit.shortcuts', lazy=False, warn=False, install=True),
    attempt_import('prompt_toolkit.history', lazy=False, warn=False, install=True),
    attempt_import('prompt_toolkit.formatted_text', lazy=False, warn=False, install=True),
    attempt_import('prompt_toolkit.styles', lazy=False, warn=False, install=True),
)
from meerschaum._internal.shell.ValidAutoSuggest import ValidAutoSuggest
from meerschaum._internal.shell.ShellCompleter import ShellCompleter
_clear_screen = get_config('shell', 'clear_screen', patch=True)
from meerschaum.utils.misc import string_width

patch = True
### remove default cmd2 commands
commands_to_remove = {
    'alias',
    'macro',
    'run_pyscript',
    'run_script',
    'shell',
    'pyscript',
    'py',
    'shortcuts',
    'history',
    'load',
}
### cmd2 only: hide commands
hidden_commands = {
    'os',
    'sh',
    'pass',
    'quit',
    'eof',
    'exit',
    '_relative_run_script',
    'ipy',
}
reserved_completers = {
    'instance', 'repo'
}

### To handle dynamic reloading, store shell attributes externally.
### This is because the shell object address gets lost upon reloads.
shell_attrs = {}

def _insert_shell_actions(
        _shell: Optional['Shell'] = None,
        actions: Optional[Dict[str, Callable[[Any], SuccessTuple]]] = None,
        keep_self: bool = False,
    ) -> None:
    """
    Update the Shell with Meerschaum actions.
    """
    from meerschaum.utils.misc import add_method_to_class
    import meerschaum._internal.shell as shell_pkg
    from meerschaum.actions import get_completer
    if actions is None:
        from meerschaum.actions import actions as _actions
        actions = _actions

    _shell_class = _shell if _shell is not None else shell_pkg.Shell

    for a, f in actions.items():
        add_method_to_class(
            func = f,
            class_def = _shell_class,
            method_name = 'do_' + a,
        )
        if a in reserved_completers:
            continue
        _completer = get_completer(a)
        if _completer is None:
            _completer = shell_pkg.default_action_completer
        completer = _completer_wrapper(_completer)
        setattr(_shell_class, 'complete_' + a, completer)

def _completer_wrapper(
        target: Callable[[Any], List[str]]
    ) -> Callable[['meerschaum._internal.shell.Shell', str, str, int, int], Any]:
    """
    Wrapper for `complete_` functions so they can instead use Meerschaum arguments.
    """
    from functools import wraps

    ### I have no idea why I had to remove `self`.
    ### Maybe it has to do with adding to an object instead of a class.
    @wraps(target)
    def wrapper(text, line, begin_index, end_index):
        _check_keys = _check_complete_keys(line)
        if _check_keys is not None:
            return _check_keys

        from meerschaum._internal.arguments._parse_arguments import parse_line
        args = parse_line(line)
        if target.__name__ != 'default_action_completer':
            if len(args['action']) > 0:
                del args['action'][0]
        args['text'] = text
        args['line'] = line
        args['begin_index'] = begin_index
        args['end_index'] = end_index
        return target(**args)

    return wrapper


def default_action_completer(
        text: Optional[str] = None,
        line: Optional[str] = None,
        begin_index: Optional[int] = None,
        end_index: Optional[int] = None,
        action: Optional[List[str]] = None,
        **kw: Any
    ) -> List[str]:
    """
    Search for subactions by default. This may be overridden by each action.
    """
    from meerschaum.actions import get_subactions
    if action is None:
        action = []
    subactions = get_subactions(action)
    sub = action[1] if len(action) > 1 else ''
    possibilities = []
    for sa in subactions:
        if sa.startswith(sub) and sa != sub:
            possibilities.append(sa)
    return sorted(possibilities)

def _check_complete_keys(line: str) -> Optional[List[str]]:
    from meerschaum._internal.arguments._parser import parser, get_arguments_triggers
    from meerschaum._internal.arguments._parse_arguments import parse_line

    ### TODO Add all triggers
    trigger_args = {
        '-c' : 'connector_keys',
        '--connector-keys' : 'connector_keys',
        '-r' : 'repository',
        '--repository' : 'repository',
        '-i' : 'mrsm_instance',
        '--instance' : 'mrsm_instance',
        '--mrsm-instance' : 'mrsm_instance',
    }


    ### TODO Find out arg possibilities
    possibilities = []
    #  last_word = line.split(' ')[-1]
    last_word = line.rstrip(' ').split(' ')[-1]

    if last_word.startswith('-'):
        for _arg, _triggers in get_arguments_triggers().items():
            for _trigger in _triggers:
                if _trigger.startswith(last_word):
                    #  and _trigger != last_word
                    if _trigger != last_word or not line.endswith(' '):
                        possibilities.append(_trigger)
        if not line.endswith(' '):
            return possibilities

    if last_word.startswith('-') and not last_word.endswith('-'):
        if line.endswith(' ') and last_word in trigger_args:
            is_trigger = True
        elif line.endswith(' '):
            ### return empty list so we don't try to parse an incomplete line.
            #  print('ABORT')
            return []

    from meerschaum.utils.misc import get_connector_labels

    if last_word.rstrip(' ') in trigger_args:
        return get_connector_labels()

    for trigger, var in trigger_args.items():
        if trigger in line:
            ### check if any text has been entered for the key
            if line.rstrip(' ').endswith(trigger):
                return get_connector_labels()

            #  args = parse_line(line.rstrip(' '))
            #  search_term = args[var] if var != 'connector_keys' else args[var][0]
            return get_connector_labels(search_term=last_word.rstrip(' '))

    return None


class Shell(cmd.Cmd):
    def __init__(
            self,
            actions: Optional[Dict[str, Any]] = None,
            sysargs: Optional[List[str]] = None
        ):
        """
        Customize the CLI from configuration
        """
        if actions is None:
            actions = {}
        if sysargs is None:
            sysargs = []
        _insert_shell_actions(_shell=self, keep_self=True)
        try:
            delattr(cmd.Cmd, '_alias_create')
            delattr(cmd.Cmd, '_alias_delete')
            delattr(cmd.Cmd, '_alias_list')
            delattr(cmd.Cmd, '_macro_create')
            delattr(cmd.Cmd, '_macro_delete')
            delattr(cmd.Cmd, '_macro_list')
        except AttributeError:
            pass

        from meerschaum.config._paths import SHELL_HISTORY_PATH
        shell_attrs['session'] = prompt_toolkit_shortcuts.PromptSession(
            history = prompt_toolkit_history.FileHistory(str(SHELL_HISTORY_PATH)),
            auto_suggest = ValidAutoSuggest(),
            completer = ShellCompleter(),
            complete_while_typing = True,
            reserve_space_for_menu = False,
        )

        try: ### try cmd2 arguments first
            super().__init__(
                allow_cli_args = False,
                auto_load_commands = False,
                persistent_history_length = 1000,
                persistent_history_file = None,
            )
            _init = True
        except Exception as e:
             ### fall back to default init (cmd)
            _init = False
        
        if not _init:
            super().__init__()

        ### remove default commands from the Cmd class
        for command in commands_to_remove:
            try:
                delattr(cmd.Cmd, f'do_{command}')
            except Exception as e:
                pass

        ### NOTE: custom actions must be added to the self._actions dictionary
        shell_attrs['_actions'] = actions
        shell_attrs['_sysargs'] = sysargs
        shell_attrs['_actions']['instance'] = self.do_instance
        shell_attrs['_actions']['repo'] = self.do_repo
        shell_attrs['_actions']['debug'] = self.do_debug
        shell_attrs['_update_bottom_toolbar'] = True
        shell_attrs['_old_bottom_toolbar'] = ''
        shell_attrs['debug'] = False
        shell_attrs['_reload'] = True
        self.load_config()
        self.hidden_commands = []
        ### update hidden commands list (cmd2 only)
        try:
            for c in hidden_commands:
                self.hidden_commands.append(c)
        except Exception as e:
            pass


    def load_config(self, instance: Optional[str] = None):
        """
        Set attributes from the shell configuration.
        """
        from meerschaum.utils.misc import remove_ansi
        from meerschaum.utils.formatting import CHARSET, ANSI, UNICODE, colored
        
        if shell_attrs.get('intro', None) != '':
            self.intro = get_config('shell', CHARSET, 'intro', patch=patch)
            self.intro += '\n' + ''.join(
                [' '
                    for i in range(
                        string_width(self.intro) - len('v' + version)
                    )
                ]
            ) + 'v' + version
        else:
            self.intro = ""
        shell_attrs['intro'] = self.intro
        shell_attrs['_prompt'] = get_config('shell', CHARSET, 'prompt', patch=patch)
        self.prompt = shell_attrs['_prompt']
        shell_attrs['ruler'] = get_config('shell', CHARSET, 'ruler', patch=patch)
        self.ruler = shell_attrs['ruler']
        shell_attrs['close_message'] = get_config('shell', CHARSET, 'close_message', patch=patch)
        self.close_message = shell_attrs['close_message']
        shell_attrs['doc_header'] = get_config('shell', CHARSET, 'doc_header', patch=patch)
        self.doc_header = shell_attrs['doc_header']
        shell_attrs['undoc_header'] = get_config('shell', CHARSET, 'undoc_header', patch=patch)
        self.undoc_header = shell_attrs['undoc_header']

        if instance is None and self.__dict__.get('instance_keys', None) is None:
            ### create default instance and repository connectors
            shell_attrs['instance_keys'] = remove_ansi(get_config('meerschaum', 'instance', patch=patch))
            ### instance is a stylized version of instance_keys
            shell_attrs['instance'] = str(shell_attrs['instance_keys'])
        else:
            shell_attrs['instance'] = instance
            shell_attrs['instance_keys'] = remove_ansi(str(instance))
        if shell_attrs.get('repo_keys', None) is None:
            shell_attrs['repo_keys'] = get_config('meerschaum', 'default_repository', patch=patch)
        ### this will be updated later in update_prompt ONLY IF {username} is in the prompt
        shell_attrs['username'] = ''

        if ANSI:
            def apply_colors(attr, key):
                return colored(
                    attr,
                    **get_config('shell', 'ansi', key, 'rich', patch=patch)
                )

            for attr_key in get_config('shell', 'ansi'):
                if attr_key not in shell_attrs:
                    continue
                shell_attrs[attr_key] = apply_colors(shell_attrs[attr_key], attr_key)
                self.__dict__[attr_key] = shell_attrs[attr_key]

        ### refresh actions
        _insert_shell_actions(_shell=self, keep_self=True)

        ### replace {instance} in prompt with stylized instance string
        self.update_prompt()

    def insert_actions(self):
        from meerschaum.actions import actions

    def update_prompt(self, instance: Optional[str] = None, username: Optional[str] = None):
        from meerschaum.utils.formatting import ANSI, colored
        from meerschaum._internal.entry import _shell, get_shell

        cmd.__builtins__['input'] = input_with_sigint(
            _old_input,
            shell_attrs['session'],
            shell = self,
        )
        prompt = shell_attrs['_prompt']
        mask = prompt
        shell_attrs['_update_bottom_toolbar'] = True

        if '{instance}' in shell_attrs['_prompt']:
            if instance is None:
                instance = shell_attrs['instance_keys']
            shell_attrs['instance'] = instance
            if ANSI:
                shell_attrs['instance'] = colored(
                    shell_attrs['instance'], **get_config(
                        'shell', 'ansi', 'instance', 'rich'
                    )
                )
            prompt = prompt.replace('{instance}', shell_attrs['instance'])
            mask = mask.replace('{instance}', ''.join(['\0' for c in '{instance}']))

        if '{username}' in shell_attrs['_prompt']:
            if username is None:
                from meerschaum.utils.misc import remove_ansi
                from meerschaum.connectors.parse import parse_instance_keys
                from meerschaum.connectors.sql import SQLConnector
                try:
                    conn_attrs = parse_instance_keys(
                        remove_ansi(shell_attrs['instance_keys']), construct=False
                    )
                    if 'username' not in conn_attrs:
                        if 'uri' in conn_attrs:
                            username = SQLConnector.parse_uri(conn_attrs['uri'])['username']
                    else:
                        username = conn_attrs['username']
                except KeyError:
                    username = '(no username)'
                except Exception as e:
                    username = str(e)
                if username is None:
                   username = '(no username)'
            shell_attrs['username'] = (
                username if not ANSI else
                colored(username, **get_config('shell', 'ansi', 'username', 'rich'))
            )
            prompt = prompt.replace('{username}', shell_attrs['username'])
            mask = mask.replace('{username}', ''.join(['\0' for c in '{username}']))

        remainder_prompt = list(shell_attrs['_prompt'])
        for i, c in enumerate(mask):
            if c != '\0':
                _c = c
                if ANSI:
                    _c = colored(_c, **get_config('shell', 'ansi', 'prompt', 'rich'))
                remainder_prompt[i] = _c
        self.prompt = ''.join(remainder_prompt).replace(
            '{username}', shell_attrs['username']
        ).replace(
            '{instance}', shell_attrs['instance']
        )
        shell_attrs['prompt'] = self.prompt
        ### flush stdout
        print("", end="", flush=True)


    def precmd(self, line: str):
        """
        Pass line string to parent actions.
        Pass parsed arguments to custom actions
        
        Overrides `default`. If an action does not exist, assume the action is `shell`
        """
        ### Preserve the working directory.
        old_cwd = os.getcwd()

        from meerschaum._internal.entry import _shell, get_shell
        self = get_shell(sysargs=shell_attrs['_sysargs'], debug=shell_attrs.get('debug', False))

        ### make a backup of line for later
        original_line = deepcopy(line)

        ### cmd2 support: check if command exists
        try:
            command = line.command
            line = str(command) + (' ' + str(line) if len(str(line)) > 0 else '')
        except Exception:
            ### we're probably running the original cmd, not cmd2
            command = None
            line = str(line)

        ### if the user specifies, clear the screen before executing any commands
        if _clear_screen:
            from meerschaum.utils.formatting._shell import clear_screen
            clear_screen(debug=shell_attrs['debug'])

        ### return blank commands (spaces break argparse)
        if original_line is None or len(str(line).strip()) == 0:
            return original_line

        if line in {
            'exit',
            'quit',
            'EOF',
        }:
            return "exit"
        ### help shortcut
        help_token = '?'
        if line.startswith(help_token):
            return "help " + line[len(help_token):]

        from meerschaum._internal.arguments import parse_line
        args = parse_line(line)
        if args.get('help', False):
            from meerschaum._internal.arguments._parser import parse_help
            parse_help(args)
            return ""

        ### NOTE: pass `shell` flag in case actions need to distinguish between
        ###       being run on the command line and being run in the shell
        args['shell'] = True
        args['line'] = line

        ### if debug is not set on the command line,
        ### default to shell setting
        if not args.get('debug', False):
            args['debug'] = shell_attrs['debug']

        ### Make sure an action was provided.
        if not args.get('action', None):
            self.emptyline()
            return ''

        ### Strip a leading 'mrsm' if it's provided.
        if args['action'][0] == 'mrsm':
            args['action'] = args['action'][1:]
            if not args['action']:
                self.emptyline()
                return ''

        ### If we don't recognize the action,
        ### make it a shell action.
        from meerschaum.actions import get_main_action_name
        main_action_name = get_main_action_name(args['action'])
        if main_action_name is None:
            if not hasattr(self, 'do_' + args['action'][0]):
                args['action'].insert(0, 'sh')
                main_action_name = 'sh'
            else:
                main_action_name = args['action'][0]

        ### if no instance is provided, use current shell default,
        ### but not for the 'api' command (to avoid recursion)
        if 'mrsm_instance' not in args and main_action_name != 'api':
            args['mrsm_instance'] = str(shell_attrs['instance_keys'])

        if 'repository' not in args and main_action_name != 'api':
            args['repository'] = str(shell_attrs['repo_keys'])

        ### parse out empty strings
        if args['action'][0].strip("\"'") == '':
            self.emptyline()
            return ""

        ### If the `--daemon` flag is present, prepend 'start job'.
        if args.get('daemon', False) and 'stack' not in args['action']:
            args['action'] = ['start', 'jobs'] + args['action']
            main_action_name = 'start'

        positional_only = (main_action_name not in shell_attrs['_actions'])
        if positional_only:
            return original_line

        from meerschaum._internal.entry import entry_with_args

        try:
            success_tuple = entry_with_args(_actions=shell_attrs['_actions'], **args)
            #  success_tuple = entry_with_args(**args)
        except Exception as e:
            success_tuple = False, str(e)

        from meerschaum.utils.formatting import print_tuple
        if isinstance(success_tuple, tuple):
            print_tuple(
                success_tuple,
                skip_common = (not shell_attrs['debug']),
                upper_padding = 1,
                lower_padding = 0,
            )

        ### Restore the old working directory.
        if old_cwd != os.getcwd():
            os.chdir(old_cwd)

        return ""

    def postcmd(self, stop : bool = False, line : str = ""):
        _reload = shell_attrs['_reload']
        if _reload:
            self.load_config(shell_attrs['instance'])
        if stop:
            return True

    def do_pass(self, line):
        """
        Do nothing.
        """

    def do_debug(self, action: Optional[List[str]] = None, **kw):
        """
        Toggle the shell's debug mode.
        If debug = on, append `--debug` to all commands.
        
        Command:
            `debug {on/true | off/false}`
            Ommitting on / off will toggle the existing value.
        """
        from meerschaum.utils.warnings import info
        on_commands = {'on', 'true', 'True'}
        off_commands = {'off', 'false', 'False'}
        if action is None:
            action = []
        try:
            state = action[0]
        except (IndexError, AttributeError):
            state = ''
        if state == '':
            shell_attrs['debug'] = not shell_attrs['debug']
        elif state.lower() in on_commands:
            shell_attrs['debug'] = True
        elif state.lower() in off_commands:
            shell_attrs['debug'] = False
        else:
            info(f"Unknown state '{state}'. Ignoring...")

        info(f"Debug mode is {'on' if shell_attrs['debug'] else 'off'}.")

    def do_instance(
            self,
            action : Optional[List[str]] = None,
            debug : bool = False,
            **kw : Any
        ) -> SuccessTuple:
        """
        Temporarily set a default Meerschaum instance for the duration of the shell.
        The default instance is loaded from the Meerschaum configuraton file
          (at keys 'meerschaum:instance').
        
        You can change the default instance with `edit config`.
        
        Usage:
            instance {instance keys}
        
        Examples:
            ```
            ### reset to default instance
            instance
        
            ### set the instance to 'api:main'
            instance api
        
            ### set the instance to a non-default connector
            instance api:myremoteinstance
            ```
        
        Note that instances must be configured, be either API or SQL connections,
          and be accessible to this machine over the network.

        """
        from meerschaum import get_connector
        from meerschaum.connectors.parse import parse_instance_keys
        from meerschaum.utils.warnings import warn, info
        from meerschaum.utils.misc import remove_ansi

        if action is None:
            action = []
        try:
            instance_keys = action[0]
        except (IndexError, AttributeError):
            instance_keys = ''
        if instance_keys == '':
            instance_keys = get_config('meerschaum', 'instance', patch=True)
        if ':' not in instance_keys:
            instance_keys += ':main'

        conn_attrs = parse_instance_keys(instance_keys, construct=False, debug=debug)
        if conn_attrs is None or not conn_attrs:
            conn_keys = str(get_connector(debug=debug))
        else:
            conn_keys = instance_keys

        shell_attrs['instance_keys'] = conn_keys

        self.update_prompt(instance=conn_keys)
        info(f"Default instance for the current shell: {conn_keys}")

        return True, "Success"


    def complete_instance(self, text: str, line: str, begin_index: int, end_index: int):
        from meerschaum.utils.misc import get_connector_labels
        from meerschaum._internal.arguments._parse_arguments import parse_line
        from meerschaum.connectors import instance_types
        args = parse_line(line)
        action = args['action']
        _text = action[1] if len(action) > 1 else ""
        return get_connector_labels(*instance_types, search_term=_text, ignore_exact_match=True)


    def do_repo(
            self,
            action: Optional[List[str]] = None,
            debug: bool = False,
            **kw: Any
        ) -> SuccessTuple:
        """
        Temporarily set a default Meerschaum repository for the duration of the shell.
        The default repository (mrsm.io) is loaded from the Meerschaum configuraton file
          (at keys 'meerschaum:default_repository').
        
        You can change the default repository with `edit config`.
        
        Usage:
            repo {API label}
        
        Examples:
            ### reset to default repository
            repo
        
            ### set the repo to 'api:main'
            repository api
        
            ### set the repository to a non-default repo
            repo myremoterepo
        
        Note that repositories are a subset of instances.
        """
        from meerschaum import get_connector
        from meerschaum.connectors.parse import parse_repo_keys
        from meerschaum.utils.warnings import warn, info

        if action is None:
            action = []

        try:
            repo_keys = action[0]
        except (IndexError, AttributeError):
            repo_keys = ''
        if repo_keys == '':
            repo_keys = get_config('meerschaum', 'default_repository', patch=True)

        conn = parse_repo_keys(repo_keys, debug=debug)
        if conn is None or not conn:
            conn = get_connector('api', debug=debug)

        shell_attrs['repo_keys'] = str(conn)

        info(f"Default repository for the current shell: {conn}")
        return True, "Success"

    def complete_repo(self, *args) -> List[str]:
        return self.complete_instance(*args)

    def do_help(self, line: str) -> List[str]:
        """
        Show help for Meerschaum actions.
        
        You can also view help for actions and subactions with `--help`.
        
        Examples:
        ```
        help show
        help show pipes
        show pipes --help
        show pipes -h
        ```
        """
        from meerschaum.actions import actions
        from meerschaum._internal.arguments._parser import parse_help
        from meerschaum._internal.arguments._parse_arguments import parse_line
        import textwrap
        args = parse_line(line)
        if len(args['action']) == 0:
            del args['action']
            shell_attrs['_actions']['show'](['actions'], **args)
            return ""
        if args['action'][0] not in shell_attrs['_actions']:
            try:
                print(textwrap.dedent(getattr(self, f"do_{args['action'][0]}").__doc__))
            except Exception as e:
                print(f"No help on '{args['action'][0]}'.")
            return ""
        parse_help(args)
        return ""

    def complete_help(self, text: str, line: str, begin_index: int, end_index: int):
        """
        Autocomplete the `help` command.
        """
        import inspect
        from meerschaum._internal.arguments._parse_arguments import parse_line
        from meerschaum.actions import get_subactions
        from meerschaum._internal.shell import Shell as _Shell
        args = parse_line(line)
        if len(args['action']) > 0 and args['action'][0] == 'help':
            ### remove 'help'
            del args['action'][0]

        action = args['action'][0] if len(args['action']) > 0 else ""
        sub = args['action'][1] if len(args['action']) > 1 else ""
        possibilities = []

        ### Search for subactions
        if sub is not None:
            for sa in get_subactions(action):
                if sa.startswith(sub) and sa != sub:
                    possibilities.append(sa)
        
        ### We found subarguments. Stop looking here.
        if len(possibilities) > 0:
            return possibilities

        ### No subactions. We're looking for an action.
        for name, f in inspect.getmembers(_Shell):
            if not inspect.isfunction(f):
                continue
            if (
                name.startswith(f'do_{action}')
                    and name != f'do_{action}'
                    and name.replace('do_', '') not in self.hidden_commands
            ):
                possibilities.append(name.replace('do_', ''))
        return possibilities

    def do_exit(self, params) -> True:
        """
        Exit the Meerschaum shell.
        """
        return True

    def emptyline(self):
        """
        If the user specifies, clear the screen.

        **NOTE:** The screen clearing is defined in the custom input below
        """

    def preloop(self):
        """
        Patch builtin cmdloop with my own input (defined below).
        """
        import signal, os
        cmd.__builtins__['input'] = input_with_sigint(
            _old_input,
            shell_attrs['session'],
            shell = self,
        )

        ### if the user specifies, clear the screen before initializing the shell
        if _clear_screen:
            from meerschaum.utils.formatting._shell import clear_screen
            clear_screen(debug=shell_attrs['debug'])

        ### if sysargs are provided, skip printing the intro and execute instead
        if shell_attrs['_sysargs']:
            shell_attrs['intro'] = ""
            self.precmd(' '.join(shell_attrs['_sysargs']))

    def postloop(self):
        print('\n' + self.close_message)

def input_with_sigint(_input, session, shell: Optional[Shell] = None):
    """
    Replace built-in `input()` with prompt_toolkit.prompt.
    """
    from meerschaum.utils.formatting import CHARSET, ANSI, UNICODE, colored
    from meerschaum.connectors import is_connected, connectors
    from meerschaum.utils.misc import remove_ansi
    from meerschaum.config import get_config
    import platform
    if shell is None:
        from meerschaum.actions import get_shell
        shell = get_shell()

    style = prompt_toolkit_styles.Style.from_dict({
        'bottom-toolbar': 'black',
    })
    last_connected = False
    def bottom_toolbar():
        nonlocal last_connected
        if not get_config('shell', 'bottom_toolbar', 'enabled'):
            return None
        if not shell_attrs['_update_bottom_toolbar'] and platform.system() == 'Windows':
            return shell_attrs['_old_bottom_toolbar']
        size = os.get_terminal_size()
        num_cols, num_lines = size.columns, size.lines

        instance_colored = (
            colored(
                shell_attrs['instance_keys'], 'on ' + get_config(
                    'shell', 'ansi', 'instance', 'rich', 'style'
                )
            ) if ANSI else colored(shell_attrs['instance_keys'], 'on white')
        )
        repo_colored = (
            colored(shell_attrs['repo_keys'], 'on ' + get_config('shell', 'ansi', 'repo', 'rich', 'style'))
            if ANSI else colored(shell_attrs['repo_keys'], 'on white')
        )
        try:
            typ, label = shell_attrs['instance_keys'].split(':')
            connected = typ in connectors and label in connectors[typ]
        except Exception as e:
            connected = False
        last_connected = connected
        connected_str = (('dis' if not connected else '') + 'connected')
        connection_text = (
            get_config(
                'formatting', connected_str, CHARSET, 'icon'
            ) + ' ' + (
                colored(connected_str.capitalize(), 'on ' + get_config(
                    'formatting', connected_str, 'ansi', 'rich', 'style'
                ) + '  ') if ANSI else (colored(connected_str.capitalize(), 'on white') + ' ')
            )
        )

        left = (
            colored(' Instance: ', 'on white') + instance_colored
            + colored('   Repo: ', 'on white') + repo_colored
        )
        right = connection_text
        buffer_size = (
            num_cols - (len(remove_ansi(left)) + len(remove_ansi(right)) + (2 if ANSI else 0))
        )
        buffer = (' ' * buffer_size) if buffer_size > 0 else '\n '
        text = left + buffer + right
        shell_attrs['_old_bottom_toolbar'] = prompt_toolkit_formatted_text.ANSI(text)
        shell_attrs['_update_bottom_toolbar'] = False
        return shell_attrs['_old_bottom_toolbar']

    def _patched_prompt(*args):
        _args = []
        for a in args:
            try:
                _a = prompt_toolkit_formatted_text.ANSI(a)
            except Exception as e:
                _a = a
            _args.append(_a)
        try:
            parsed = session.prompt(*_args, bottom_toolbar=bottom_toolbar, style=style)
            ### clear screen on empty input
            ### NOTE: would it be better to do nothing instead?
            if len(parsed.strip()) == 0:
                if _clear_screen:
                    from meerschaum.utils.formatting._shell import clear_screen
                    clear_screen()
        except KeyboardInterrupt:
            print("^C")
            return "pass"
        return parsed

    return _patched_prompt



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Define SQLAlchemy tables
"""

from __future__ import annotations
from meerschaum.utils.typing import Optional, Dict, Union, InstanceConnector, List
from meerschaum.utils.warnings import error, warn

### store a tables dict for each connector
connector_tables = {}

_sequence_flavors = {'duckdb', 'oracle'}
_skip_index_names_flavors = {'mssql',}

def get_tables(
        mrsm_instance: Optional[Union[str, InstanceConnector]] = None,
        create: bool = True,
        debug: Optional[bool] = None
    ) -> Union[Dict[str, 'sqlalchemy.Table'], bool]:
    """
    Create tables on the database and return the `sqlalchemy` tables.

    Parameters
    ----------
    mrsm_instance: Optional[Union[str, InstanceConnector]], default None
        The connector on which the tables reside.

    create: bool, default True:
        If `True`, create the tables if they don't exist.

    debug: Optional[bool], default None:
        Verbosity Toggle.

    Returns
    -------
    A dictionary of `sqlalchemy.Table` objects if the connector is a `SQLConnector`.
    Otherwise just a bool for `APIConnector` objects.

    """
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.formatting import pprint
    from meerschaum.connectors.parse import parse_instance_keys
    from meerschaum.utils.packages import attempt_import
    from meerschaum.utils.sql import json_flavors
    from meerschaum import get_connector

    sqlalchemy, sqlalchemy_dialects_postgresql = attempt_import(
        'sqlalchemy',
        'sqlalchemy.dialects.postgresql',
        lazy = False
    )
    if not sqlalchemy:
        error(f"Failed to import sqlalchemy. Is sqlalchemy installed?")

    if mrsm_instance is None:
        conn = get_connector(debug=debug)
    elif isinstance(mrsm_instance, str):
        conn = parse_instance_keys(mrsm_instance, debug=debug)
    else: ### NOTE: mrsm_instance MUST BE a SQL Connector for this to work!
        conn = mrsm_instance

    ### kind of a hack. Create the tables remotely
    from meerschaum.connectors.api import APIConnector
    if isinstance(conn, APIConnector):
        if create:
            return conn.create_metadata(debug=debug)
        return {}

    ### Skip if the connector is not a SQL connector.
    if getattr(conn, 'type', None) != 'sql':
        return {}

    if conn not in connector_tables:
        if debug:
            dprint(f"Creating tables for connector '{conn}'.")

        id_type = sqlalchemy.Integer
        if conn.flavor in json_flavors:
            params_type = sqlalchemy.types.JSON
        else:
            params_type = sqlalchemy.types.Text
        id_names = ('user_id', 'plugin_id', 'pipe_id')
        sequences = {
            k: sqlalchemy.Sequence(k + '_seq')
            for k in id_names 
        }
        id_col_args = { k: [k, id_type] for k in id_names }
        id_col_kw = { k: {'primary_key': True} for k in id_names }
        index_names = conn.flavor not in _skip_index_names_flavors

        if conn.flavor in _sequence_flavors:
            for k, args in id_col_args.items():
                args.append(sequences[k])
            for k, kw in id_col_kw.items():
                kw.update({'server_default': sequences[k].next_value()})

        _tables = {
            'users': sqlalchemy.Table(
                'mrsm_users',
                conn.metadata,
                sqlalchemy.Column(
                    *id_col_args['user_id'],
                    **id_col_kw['user_id'],
                ),
                sqlalchemy.Column(
                    'username',
                    sqlalchemy.String(256),
                    index = index_names,
                    nullable = False,
                ),
                sqlalchemy.Column('password_hash', sqlalchemy.String(1024)),
                sqlalchemy.Column('email', sqlalchemy.String(256)),
                sqlalchemy.Column('user_type', sqlalchemy.String(256)),
                sqlalchemy.Column('attributes', params_type),
                extend_existing = True,
            ),
            'plugins': sqlalchemy.Table(
                *([
                    'mrsm_plugins',
                    conn.metadata,
                    sqlalchemy.Column(
                        *id_col_args['plugin_id'],
                        **id_col_kw['plugin_id'],
                    ),
                    sqlalchemy.Column(
                        'plugin_name', sqlalchemy.String(256), index=index_names, nullable=False,
                    ),
                    sqlalchemy.Column('user_id', sqlalchemy.Integer, nullable=False),
                    sqlalchemy.Column('version', sqlalchemy.String(256)),
                    sqlalchemy.Column('attributes', params_type),
                ] + ([
                    sqlalchemy.ForeignKeyConstraint(['user_id'], ['mrsm_users.user_id']),
                ] if conn.flavor != 'duckdb' else [])),
                extend_existing = True,
            ),
            'temp_tables': sqlalchemy.Table(
                'mrsm_temp_tables',
                conn.metadata,
                sqlalchemy.Column(
                    'date_created',
                    sqlalchemy.DateTime,
                    index = True,
                    nullable = False,
                ),
                sqlalchemy.Column(
                    'table',
                    sqlalchemy.String(256),
                    index = index_names,
                    nullable = False,
                ),
                sqlalchemy.Column(
                    'ready_to_drop',
                    sqlalchemy.DateTime,
                    index = False,
                    nullable = True,
                ),
                extend_existing = True,
            ),
        }

        _tables['pipes'] = sqlalchemy.Table(
            "mrsm_pipes",
            conn.metadata,
            sqlalchemy.Column(
                *id_col_args['pipe_id'],
                **id_col_kw['pipe_id'],
            ),
            sqlalchemy.Column(
                "connector_keys",
                sqlalchemy.String(256),
                index = index_names,
                nullable = False,
            ),
            sqlalchemy.Column(
                "metric_key",
                sqlalchemy.String(256),
                index = index_names,
                nullable = False,
            ),
            sqlalchemy.Column(
                "location_key",
                sqlalchemy.String(256),
                index = index_names,
                nullable = True,
            ),
            sqlalchemy.Column("parameters", params_type),
            extend_existing = True,
        )

        ### store the table dict for reuse (per connector)
        connector_tables[conn] = _tables
        if create:
            create_schemas(
                conn,
                schemas = [conn.internal_schema],
                debug = debug,
            )
            create_tables(conn, tables=_tables)

    return connector_tables[conn]


def create_tables(
        conn: 'meerschaum.connectors.SQLConnector',
        tables: Optional[Dict[str, 'sqlalchemy.Table']] = None,
    ) -> bool:
    """
    Create the tables on the database.
    """
    from meerschaum.utils.sql import get_rename_table_queries, table_exists
    _tables = tables if tables is not None else get_tables(conn)

    try:
        conn.metadata.create_all(bind=conn.engine)
    except Exception as e:
        import traceback
        traceback.print_exc()
        warn(str(e))
        return False
    return True


def create_schemas(
        conn: 'meerschaum.connectors.SQLConnector',
        schemas: List[str],
        debug: bool = False,
    ) -> bool:
    """
    Create the internal Meerschaum schema on the database.
    """
    from meerschaum.config.static import STATIC_CONFIG
    from meerschaum.utils.packages import attempt_import
    from meerschaum.utils.sql import sql_item_name, NO_SCHEMA_FLAVORS, SKIP_IF_EXISTS_FLAVORS
    if conn.flavor in NO_SCHEMA_FLAVORS:
        return True

    sqlalchemy_schema = attempt_import('sqlalchemy.schema')
    successes = {}
    skip_if_not_exists = conn.flavor in SKIP_IF_EXISTS_FLAVORS
    if_not_exists_str = ("IF NOT EXISTS " if not skip_if_not_exists else "")
    with conn.engine.connect() as connection:
        for schema in schemas:
            if not schema:
                continue
            schema_name = sql_item_name(schema, conn.flavor)
            schema_exists = conn.engine.dialect.has_schema(connection, schema)
            if schema_exists:
                continue

            create_schema_query = f"CREATE SCHEMA {if_not_exists_str}{schema_name}"
            try:
                result = conn.exec(create_schema_query, debug=debug)
                successes[schema] = (result is not None)
            except Exception as e:
                warn(f"Failed to create internal schema '{schema}':\n{e}")
    return all(successes.values())



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Define the parent `Connector` class.
"""

from __future__ import annotations
import abc
import copy
from meerschaum.utils.typing import Iterable, Optional, Any, Union, List, Dict

class InvalidAttributesError(Exception):
    """
    Raised when the incorrect attributes are set in the Connector.
    """

class Connector(metaclass=abc.ABCMeta):
    """
    The base connector class to hold connection attributes,
    """
    def __init__(
            self,
            type: Optional[str] = None,
            label: Optional[str] = None,
            **kw: Any
        ):
        """
        Parameters
        ----------
        type: str
            The `type` of the connector (e.g. `sql`, `api`, `plugin`).

        label: str
            The `label` for the connector.

        Run `mrsm edit config` and to edit connectors in the YAML file:

        ```
        meerschaum:
            connections:
                {type}:
                    {label}:
                        ### attributes go here
        ```

        """
        self._original_dict = copy.deepcopy(self.__dict__)
        self._set_attributes(type=type, label=label, **kw)
        self.verify_attributes(getattr(self, 'REQUIRED_ATTRIBUTES', None))

    def _reset_attributes(self):
        self.__dict__ = self._original_dict

    def _set_attributes(
            self,
            *args,
            inherit_default: bool = True,
            **kw: Any
        ):
        from meerschaum.config.static import STATIC_CONFIG
        from meerschaum.utils.warnings import error

        self._attributes = {}

        default_label = STATIC_CONFIG['connectors']['default_label']

        ### NOTE: Support the legacy method of explicitly passing the type.
        label = kw.get('label', None)
        if label is None:
            if len(args) == 2:
                label = args[1]
            elif len(args) == 0:
                label = None
            else:
                label = args[0]

        if label == 'default':
            error(
                f"Label cannot be 'default'. Did you mean '{default_label}'?",
                InvalidAttributesError,
            )
        self.__dict__['label'] = label

        from meerschaum.config import get_config
        conn_configs = copy.deepcopy(get_config('meerschaum', 'connectors'))
        connector_config = copy.deepcopy(get_config('system', 'connectors'))

        ### inherit attributes from 'default' if exists
        if inherit_default:
            inherit_from = 'default'
            if self.type in conn_configs and inherit_from in conn_configs[self.type]:
                _inherit_dict = copy.deepcopy(conn_configs[self.type][inherit_from])
                self._attributes.update(_inherit_dict)

        ### load user config into self._attributes
        if self.type in conn_configs and self.label in conn_configs[self.type]:
            self._attributes.update(conn_configs[self.type][self.label])

        ### load system config into self._sys_config
        ### (deep copy so future Connectors don't inherit changes)
        if self.type in connector_config:
            self._sys_config = copy.deepcopy(connector_config[self.type])

        ### add additional arguments or override configuration
        self._attributes.update(kw)

        ### finally, update __dict__ with _attributes.
        self.__dict__.update(self._attributes)


    def verify_attributes(
            self,
            required_attributes: Optional[List[str]] = None,
            debug: bool = False
        ) -> None:
        """
        Ensure that the required attributes have been met.
        
        The Connector base class checks the minimum requirements.
        Child classes may enforce additional requirements.

        Parameters
        ----------
        required_attributes: Optional[List[str]], default None
            Attributes to be verified. If `None`, default to `['label']`.

        debug: bool, default False
            Verbosity toggle.

        Returns
        -------
        Don't return anything.

        Raises
        ------
        An error if any of the required attributes are missing.
        """
        from meerschaum.utils.warnings import error, warn
        from meerschaum.utils.debug import dprint
        from meerschaum.utils.misc import items_str
        if required_attributes is None:
            required_attributes = ['label']
        missing_attributes = set()
        for a in required_attributes:
            if a not in self.__dict__:
                missing_attributes.add(a)
        if len(missing_attributes) > 0:
            error(
                (
                    f"Missing {items_str(list(missing_attributes))} "
                    + f"for connector '{self.type}:{self.label}'."
                ),
                InvalidAttributesError,
                silent = True,
                stack = False
            )


    def __str__(self):
        """
        When cast to a string, return type:label.
        """
        return f"{self.type}:{self.label}"

    def __repr__(self):
        """
        Represent the connector as type:label.
        """
        return str(self)

    @property
    def meta(self) -> Dict[str, Any]:
        """
        Return the keys needed to reconstruct this Connector.
        """
        _meta = {
            key: value
            for key, value in self.__dict__.items()
            if not str(key).startswith('_')
        }
        _meta.update({
            'type': self.type,
            'label': self.label,
        })
        return _meta


    @property
    def type(self) -> str:
        """
        Return the type for this connector.
        """
        _type = self.__dict__.get('type', None)
        if _type is None:
            import re
            _type = re.sub(r'connector$', '', self.__class__.__name__.lower())
            self.__dict__['type'] = _type
        return _type


    @property
    def label(self) -> str:
        """
        Return the label for this connector.
        """
        _label = self.__dict__.get('label', None)
        if _label is None:
            from meerschaum.config.static import STATIC_CONFIG
            _label = STATIC_CONFIG['connectors']['default_label']
            self.__dict__['label'] = _label
        return _label




#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Utility functions pertaining to network functionality (e.g. available ports).
"""

def is_port_in_use(port: int):
    """Try to bind to a port and return whether it is available.

    Parameters
    ----------
    port: int :
        

    Returns
    -------

    """
    import socket
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(('localhost', port)) == 0


def find_open_ports(start_port: int = 8000, end_port: int = 9001):
    """Check a range of ports and yield open ports.
    
    Found from StackExchange here:
    https://codereview.stackexchange.com/questions/116450/find-available-ports-on-localhost

    Parameters
    ----------
    start_port: int :
         (Default value = 8000)
    end_port: int :
         (Default value = 9001)

    Returns
    -------

    """
    for port in range(start_port, end_port):
        if not is_port_in_use(port):
            yield port

